""" Loading Raw data """
import logging
from dataclasses import dataclass
from numbers import Number
from pathlib import Path
from typing import List, Tuple, Union

import numpy as np
import pandas as pd
import xarray as xr

import nowcasting_dataset.filesystem.utils as nd_fs_utils
from nowcasting_dataset.data_sources.data_source import DataSource
from nowcasting_dataset.data_sources.sun.raw_data_load_save import load_from_zarr, x_y_to_name
from nowcasting_dataset.data_sources.sun.sun_model import Sun
from nowcasting_dataset.geospatial import calculate_azimuth_and_elevation_angle

logger = logging.getLogger(__name__)


@dataclass
class SunDataSource(DataSource):
    """Add azimuth and elevation angles of the sun."""

    zarr_path: Union[str, Path]

    def __post_init__(self):
        """Post Init"""
        super().__post_init__()
        self._load()

    @staticmethod
    def get_data_model_for_batch():
        """Get the model that is used in the batch"""
        return Sun

    def check_input_paths_exist(self) -> None:
        """Check input paths exist.  If not, raise a FileNotFoundError."""
        nd_fs_utils.check_path_exists(self.zarr_path)

    def get_example(
        self, t0_dt: pd.Timestamp, x_meters_center: Number, y_meters_center: Number
    ) -> xr.Dataset:
        """
        Get example data from t0_dt and x and y xoordinates

        Args:
            t0_dt: the timestamp to get the sun data for
            x_meters_center: the x coordinate (OSGB)
            y_meters_center: the y coordinate (OSGB)

        Returns: Dictionary of azimuth and elevation data
        """
        # all sun data is from 2019, analaysis showed over the timescale we are interested in the
        # elevation and azimuth angles change by < 1 degree, so to save data, we just use data
        # from 2019.
        t0_dt = t0_dt.replace(year=2019)

        start_dt = self._get_start_dt(t0_dt)
        end_dt = self._get_end_dt(t0_dt)

        # The names of the columns get truncated when saving, therefore we need to look for the
        # name of the columns near the location we are looking for
        locations = np.array(
            [[float(z.split(",")[0]), float(z.split(",")[1])] for z in self.azimuth.columns]
        )
        location = locations[
            np.isclose(locations[:, 0], x_meters_center)
            & np.isclose(locations[:, 1], y_meters_center)
        ]
        # lets make sure there is atleast one
        assert len(location) > 0
        # Take the first location, and x and y coordinates are the first and center entries in
        # this array.
        location = location[0]
        # make name of column to pull data from. The columns name will be about
        # something like '22222.555,3333.6666'
        name = x_y_to_name(x=location[0], y=location[1])

        del x_meters_center, y_meters_center
        azimuth = self.azimuth.loc[start_dt:end_dt][name]
        elevation = self.elevation.loc[start_dt:end_dt][name]

        azimuth = azimuth.to_xarray().rename({"index": "time"})
        elevation = elevation.to_xarray().rename({"index": "time"})

        sun = azimuth.to_dataset(name="azimuth")
        sun["elevation"] = elevation

        return sun

    def _load(self):

        logger.info(f"Loading Sun data from {self.zarr_path}")

        self.azimuth, self.elevation = load_from_zarr(zarr_path=self.zarr_path)

    def get_locations(self, t0_datetimes: pd.DatetimeIndex) -> Tuple[List[Number], List[Number]]:
        """Sun data should not be used to get batch locations"""
        raise NotImplementedError("Sun data should not be used to get batch locations")

    def datetime_index(self) -> pd.DatetimeIndex:
        """Get datetimes where elevation >= 10"""

        # get the lat and lon from london
        latitude = 51
        longitude = 0

        # get elevation for all datetimes
        azimuth_elevation = calculate_azimuth_and_elevation_angle(
            latitude=latitude, longitude=longitude, datestamps=self.elevation.index
        )

        # only select elevations > 10
        mask = azimuth_elevation["elevation"] >= 10

        # create warnings, so we know how many datetimes will be dropped.
        # Should be slightly more than half as its night time 50% of the time
        n_dropping = len(azimuth_elevation) - sum(mask)
        logger.debug(
            f"Will be dropping {n_dropping} datetimes "
            f"out of {len(azimuth_elevation)} as elevation is < 10"
        )

        datetimes = self.elevation[mask].index

        # Sun data is only for 2019, so to expand on these by
        # repeating data from 2014 to 2023
        all_datetimes = pd.DatetimeIndex([])
        for delta_years in range(-5, 5, 1):
            on_year = datetimes + pd.offsets.DateOffset(months=12 * delta_years)
            all_datetimes = all_datetimes.append(on_year)

        return all_datetimes



"""Manager class."""

import logging
import multiprocessing
from functools import partial
from pathlib import Path
from typing import Optional, Union

import numpy as np
import pandas as pd

# nowcasting_dataset imports
import nowcasting_dataset.time as nd_time
import nowcasting_dataset.utils as nd_utils
from nowcasting_dataset import config
from nowcasting_dataset.consts import (
    SPATIAL_AND_TEMPORAL_LOCATIONS_COLUMN_NAMES,
    SPATIAL_AND_TEMPORAL_LOCATIONS_OF_EACH_EXAMPLE_FILENAME,
)
from nowcasting_dataset.data_sources import ALL_DATA_SOURCE_NAMES, MAP_DATA_SOURCE_NAME_TO_CLASS
from nowcasting_dataset.dataset.split import split
from nowcasting_dataset.filesystem import utils as nd_fs_utils

logger = logging.getLogger(__name__)


class Manager:
    """The Manager initialises and manage a dict of DataSource objects.

    Attrs:
      config: Configuration object.
      data_sources: dict[str, DataSource]
      data_source_which_defines_geospatial_locations: DataSource: The DataSource used to compute the
        geospatial locations of each example.
      save_batches_locally_and_upload: bool: Set to True by `load_yaml_configuration()` if
        `config.process.upload_every_n_batches > 0`.
    """

    def __init__(self) -> None:  # noqa: D107
        self.config = None
        self.data_sources = {}
        self.data_source_which_defines_geospatial_locations = None

    def load_yaml_configuration(self, filename: str) -> None:
        """Load YAML config from `filename`."""
        logger.debug(f"Loading YAML configuration file {filename}")
        self.config = config.load_yaml_configuration(filename)
        self.config = config.set_git_commit(self.config)
        self.save_batches_locally_and_upload = self.config.process.upload_every_n_batches > 0
        logger.debug(f"config={self.config}")

    def save_yaml_configuration(self):
        """Save configuration to the 'output_data' location"""
        config.save_yaml_configuration(configuration=self.config)

    def configure_loggers(
        self,
        log_level: str,
        names_of_selected_data_sources: Optional[list[str]] = ALL_DATA_SOURCE_NAMES,
    ) -> None:
        """Configure loggers.

        Print combined log to stdout.
        Save combined log to self.config.output_data.filepath / combined.log
        Save individual logs for each DataSource in
            self.config.output_data.filepath / <data_source>.log
        """

        # make sure the folder exists
        nd_fs_utils.makedirs(path=self.config.output_data.filepath)

        # Configure combined logger.
        combined_log_filename = self.config.output_data.filepath / "combined.log"
        nd_utils.configure_logger(
            log_level=log_level,
            logger_name="nowcasting_dataset",
            handlers=[
                logging.StreamHandler(),
                logging.FileHandler(combined_log_filename, mode="a"),
            ],
        )

        # Configure loggers for each DataSource.
        for data_source_name in names_of_selected_data_sources:
            log_filename = self.config.output_data.filepath / f"{data_source_name}.log"
            nd_utils.configure_logger(
                log_level=log_level,
                logger_name=f"nowcasting_dataset.data_sources.{data_source_name}",
                handlers=[logging.FileHandler(log_filename, mode="a")],
            )

    def initialise_data_sources(
        self, names_of_selected_data_sources: Optional[list[str]] = ALL_DATA_SOURCE_NAMES
    ) -> None:
        """Initialize DataSources specified in the InputData configuration.

        For each key in each DataSource's configuration object, the string `<data_source_name>_`
        is removed from the key before passing to the DataSource constructor.  This allows us to
        have verbose field names in the configuration YAML files, whilst also using standard
        constructor arguments for DataSources.
        """
        for data_source_name in names_of_selected_data_sources:
            logger.debug(f"Creating {data_source_name} DataSource object.")
            config_for_data_source = getattr(self.config.input_data, data_source_name)
            if config_for_data_source is None:
                logger.info(f"No configuration found for {data_source_name}.")
                continue
            config_for_data_source = config_for_data_source.dict()

            # Strip `<data_source_name>_` from the config option field names.
            config_for_data_source = nd_utils.remove_regex_pattern_from_keys(
                config_for_data_source, pattern_to_remove=f"^{data_source_name}_"
            )

            data_source_class = MAP_DATA_SOURCE_NAME_TO_CLASS[data_source_name]
            try:
                data_source = data_source_class(**config_for_data_source)
            except Exception:
                logger.exception(f"Exception whilst instantiating {data_source_name}!")
                raise
            self.data_sources[data_source_name] = data_source

        # Set data_source_which_defines_geospatial_locations:
        try:
            self.data_source_which_defines_geospatial_locations = self.data_sources[
                self.config.input_data.data_source_which_defines_geospatial_locations
            ]
        except KeyError:
            if self._locations_csv_file_exists():
                logger.info(
                    f"{self.config.input_data.data_source_which_defines_geospatial_locations=}"
                    " is not a member of the DataSources, but that does not matter because the CSV"
                    " files which specify the locations of the examples already exists!"
                )
            else:
                msg = (
                    "input_data.data_source_which_defines_geospatial_locations="
                    f"{self.config.input_data.data_source_which_defines_geospatial_locations}"
                    " is not a member of the DataSources, so cannot set"
                    " self.data_source_which_defines_geospatial_locations!"
                    f" The available DataSources are: {list(self.data_sources.keys())}"
                )
                logger.error(msg)
                raise RuntimeError(msg)
        else:
            logger.info(
                f"DataSource `{data_source_name}` set as"
                " data_source_which_defines_geospatial_locations."
            )

    def create_files_specifying_spatial_and_temporal_locations_of_each_example_if_necessary(
        self,
    ) -> None:
        """Creates CSV files specifying the locations of each example if those files don't exist yet.

        Creates one file per split, in this location:

        `<output_data.filepath> / <split_name> / spatial_and_temporal_locations_of_each_example.csv`

        Creates the output directory if it does not exist.

        Works on any compute environment.
        """
        if self._locations_csv_file_exists():
            logger.info(
                f"{SPATIAL_AND_TEMPORAL_LOCATIONS_OF_EACH_EXAMPLE_FILENAME} already exists!"
            )
            return
        logger.info(
            f"{SPATIAL_AND_TEMPORAL_LOCATIONS_OF_EACH_EXAMPLE_FILENAME} does not exist so"
            " will create..."
        )
        t0_datetimes = self.get_t0_datetimes_across_all_data_sources(
            freq=self.config.process.t0_datetime_frequency
        )
        # TODO: move hard code values to config file #426
        split_t0_datetimes = split.split_data(
            datetimes=t0_datetimes,
            method=self.config.process.split_method,
            train_test_validation_split=self.config.process.train_test_validation_split,
            train_validation_test_datetime_split=[
                pd.Timestamp("2020-01-01"),
                pd.Timestamp("2021-01-01"),
            ],
        )
        for split_name, datetimes_for_split in split_t0_datetimes._asdict().items():
            path_for_csv = self.config.output_data.filepath / split_name
            n_batches_requested = self._get_n_batches_requested_for_split_name(split_name)
            if (n_batches_requested == 0 and len(datetimes_for_split) != 0) or (
                len(datetimes_for_split) == 0 and n_batches_requested != 0
            ):
                # TODO: Issue #450: Test this scenario!
                msg = (
                    f"For split {split_name}: n_{split_name}_batches={n_batches_requested} and"
                    f" {len(datetimes_for_split)=}!  This is an error!"
                    f"  If n_{split_name}_batches==0 then len(datetimes_for_split) must also"
                    f" equal 0, and visa-versa!  Please check `n_{split_name}_batches` and"
                    " `split_method` in the config YAML!"
                )
                logger.error(msg)
                raise RuntimeError(msg)
            if n_batches_requested == 0:
                logger.info(f"0 batches requested for {split_name} so won't create {path_for_csv}")
                continue
            n_examples = n_batches_requested * self.config.process.batch_size
            logger.debug(
                f"Creating {n_batches_requested:,d} batches x {self.config.process.batch_size:,d}"
                f" examples per batch = {n_examples:,d} examples for {split_name}."
            )

            # for the test set, we want to get all locations for each datetime,
            # for the train and validation set, we just want one location for each datetime
            if split_name == split.SplitName.TEST.value:
                get_all_locations = True
            else:
                get_all_locations = False

            df_of_locations = self.sample_spatial_and_temporal_locations_for_examples(
                t0_datetimes=datetimes_for_split,
                n_examples=n_examples,
                get_all_locations=get_all_locations,
            )
            output_filename = self._filename_of_locations_csv_file(split_name)
            logger.info(f"Making {path_for_csv} if it does not exist.")
            nd_fs_utils.makedirs(path_for_csv, exist_ok=True)
            logger.debug(f"Writing {output_filename}")
            df_of_locations.to_csv(output_filename)

    def _get_n_batches_requested_for_split_name(self, split_name: str) -> int:
        return getattr(self.config.process, f"n_{split_name}_batches")

    def _filename_of_locations_csv_file(self, split_name: str) -> Path:
        return (
            self.config.output_data.filepath
            / split_name
            / SPATIAL_AND_TEMPORAL_LOCATIONS_OF_EACH_EXAMPLE_FILENAME
        )

    def _locations_csv_file_exists(self) -> bool:
        """Check if filepath/train/spatial_and_temporal_locations_of_each_example.csv exists."""
        filename = self._filename_of_locations_csv_file(split_name=split.SplitName.TRAIN.value)
        try:
            nd_fs_utils.check_path_exists(filename)
        except FileNotFoundError:
            logger.info(f"{filename} does not exist!")
            return False
        else:
            logger.info(f"{filename} exists!")
            return True

    def get_t0_datetimes_across_all_data_sources(
        self, freq: Union[str, pd.Timedelta]
    ) -> pd.DatetimeIndex:
        """
        Compute the intersection of the t0 datetimes available across all DataSources.

        Args:
            freq: The Pandas frequency string. The returned DatetimeIndex will be at this frequency,
                and every datetime will be aligned to this frequency.  For example, if
                freq='5 minutes' then every datetime will be at 00, 05, ..., 55 minutes
                past the hour.

        Returns:  Valid t0 datetimes, taking into consideration all DataSources,
            filtered by daylight hours (SatelliteDataSource.datetime_index() removes the night
            datetimes).
        """
        logger.debug(
            f"Getting the intersection of time periods across all DataSources at freq={freq}..."
        )
        if set(self.data_sources.keys()) != set(ALL_DATA_SOURCE_NAMES):
            logger.warning(
                "Computing available t0 datetimes using less than all available DataSources!"
                " Are you sure you mean to do this?!"
            )

        # Get the intersection of t0 time periods from all data sources.
        t0_time_periods_for_all_data_sources = []
        for data_source_name, data_source in self.data_sources.items():
            logger.debug(f"Getting t0 time periods for {data_source_name}")
            try:
                t0_time_periods = data_source.get_contiguous_t0_time_periods()
            except NotImplementedError:
                # Skip data_sources with no concept of time.
                logger.debug(f"Skipping {data_source_name} because it has not concept of datetime.")
            else:
                t0_time_periods_for_all_data_sources.append(t0_time_periods)

        intersection_of_t0_time_periods = nd_time.intersection_of_multiple_dataframes_of_periods(
            t0_time_periods_for_all_data_sources
        )

        t0_datetimes = nd_time.time_periods_to_datetime_index(
            time_periods=intersection_of_t0_time_periods, freq=freq
        )
        logger.debug(
            f"Found {len(t0_datetimes):,d} datetimes at freq=`{freq}` across"
            f" DataSources={list(self.data_sources.keys())}."
            f"  From {t0_datetimes[0]} to {t0_datetimes[-1]}."
        )
        return t0_datetimes

    def sample_spatial_and_temporal_locations_for_examples(
        self, t0_datetimes: pd.DatetimeIndex, n_examples: int, get_all_locations: bool = False
    ) -> pd.DataFrame:
        """
        Computes the geospatial and temporal locations for each training example.

        The first data_source in this DataSourceList defines the geospatial locations of
        each example.

        Args:
            t0_datetimes: All available t0 datetimes.  Can be computed with
                `DataSourceList.get_t0_datetimes_across_all_data_sources()`
            n_examples: The number of examples requested.
            get_all_locations: optional to return all locations for each t0_datetime

        Returns:
            Each row of each the DataFrame specifies the position of each example, using
            columns: 't0_datetime_UTC', 'x_center_OSGB', 'y_center_OSGB'.
        """
        assert len(t0_datetimes) > 0
        assert type(t0_datetimes) == pd.DatetimeIndex

        if get_all_locations:
            # Because we are going to get all locations for each datetime,
            # lets divided the number of examples by the number of locations in the data source
            number_locations = (
                self.data_source_which_defines_geospatial_locations.get_number_locations()
            )
            assert n_examples % number_locations == 0, (
                f"{n_examples=} needs to be a multiple of {number_locations=} "
                f"because we are getting an example for each location"
            )
            n_examples = int(n_examples / number_locations)

        shuffled_t0_datetimes = np.random.choice(t0_datetimes, size=n_examples)
        # TODO: Issue #304. Speed this up by splitting the shuffled_t0_datetimes across
        # multiple processors.  Currently takes about half an hour for 25,000 batches.
        # But wait until we've implemented issue #305, as that is likely to be sufficient!

        # make sure 'shuffled_t0_datetimes' is pd.DatetimeIndex
        shuffled_t0_datetimes = pd.DatetimeIndex(shuffled_t0_datetimes)

        if get_all_locations:

            # note that the returned 'shuffled_t0_datetimes'
            # has duplicate datetimes for each location
            (
                shuffled_t0_datetimes,
                x_locations,
                y_locations,
            ) = self.data_source_which_defines_geospatial_locations.get_all_locations(
                t0_datetimes_utc=shuffled_t0_datetimes
            )

        else:

            (
                x_locations,
                y_locations,
            ) = self.data_source_which_defines_geospatial_locations.get_locations(
                shuffled_t0_datetimes
            )

        return pd.DataFrame(
            {
                "t0_datetime_UTC": shuffled_t0_datetimes,
                "x_center_OSGB": x_locations,
                "y_center_OSGB": y_locations,
            }
        )

    def _get_first_batches_to_create(
        self, overwrite_batches: bool
    ) -> dict[split.SplitName, dict[str, int]]:
        """For each SplitName & for each DataSource name, return the first batch ID to create.

        For example, the returned_dict[SplitName.TRAIN]['gsp'] tells us the first batch_idx to
        create for the training set for the GSPDataSource.
        """
        # Initialise to zero:
        first_batches_to_create: dict[split.SplitName, dict[str, int]] = {}
        for split_name in split.SplitName:
            first_batches_to_create[split_name] = {
                data_source_name: 0 for data_source_name in self.data_sources
            }

        if overwrite_batches:
            return first_batches_to_create

        # If we're not overwriting batches then find the last batch on disk.
        for split_name in split.SplitName:
            for data_source_name in self.data_sources:
                path = (
                    self.config.output_data.filepath / split_name.value / data_source_name / "*.nc"
                )
                try:
                    max_batch_id_on_disk = nd_fs_utils.get_maximum_batch_id(path)
                except FileNotFoundError:
                    max_batch_id_on_disk = -1
                first_batches_to_create[split_name][data_source_name] = max_batch_id_on_disk + 1

        return first_batches_to_create

    def _check_if_more_batches_are_required_for_split(
        self,
        split_name: split.SplitName,
        first_batches_to_create: dict[split.SplitName, dict[str, int]],
    ) -> bool:
        """Returns True if batches still need to be created for any DataSource."""
        n_batches_requested = self._get_n_batches_requested_for_split_name(split_name.value)
        for data_source_name in self.data_sources:
            if first_batches_to_create[split_name][data_source_name] < n_batches_requested:
                return True
        return False

    def _find_splits_which_need_more_batches(
        self,
        first_batches_to_create: dict[split.SplitName, dict[str, int]],
    ) -> list[split.SplitName]:
        """Returns list of SplitNames which need more batches to be produced."""
        return [
            split_name
            for split_name in split.SplitName
            if self._check_if_more_batches_are_required_for_split(
                split_name=split_name,
                first_batches_to_create=first_batches_to_create,
            )
        ]

    def create_batches(self, overwrite_batches: bool) -> None:
        """Create batches (if necessary).

        Make dirs: `<output_data.filepath> / <split_name> / <data_source_name>`.

        Also make `local_temp_path` if necessary.

        Args:
          overwrite_batches: If True then start from batch 0, regardless of which batches have
            previously been written to disk. If False then check which batches have previously been
            written to disk, and only create any batches which have not yet been written to disk.
        """
        logger.debug("Entering Manager.create_batches...")
        first_batches_to_create = self._get_first_batches_to_create(overwrite_batches)

        # Check if there's any work to do.
        if overwrite_batches:
            splits_which_need_more_batches = [
                split_name
                for split_name in split.SplitName
                if self._get_n_batches_requested_for_split_name(split_name.value) > 0
            ]
        else:
            splits_which_need_more_batches = self._find_splits_which_need_more_batches(
                first_batches_to_create=first_batches_to_create
            )
            if len(splits_which_need_more_batches) == 0:
                logger.info("All batches have already been created!  No work to do!")
                return

        # Load locations for each example off disk.
        locations_for_each_example_of_each_split: dict[split.SplitName, pd.DataFrame] = {}
        for split_name in splits_which_need_more_batches:
            filename = self._filename_of_locations_csv_file(split_name.value)
            logger.info(f"Loading {filename}.")
            locations_for_each_example = pd.read_csv(filename, index_col=0)
            assert locations_for_each_example.columns.to_list() == list(
                SPATIAL_AND_TEMPORAL_LOCATIONS_COLUMN_NAMES
            )
            # Converting to datetimes is much faster using `pd.to_datetime()` than
            # passing `parse_datetimes` into `pd.read_csv()`.
            locations_for_each_example["t0_datetime_UTC"] = pd.to_datetime(
                locations_for_each_example["t0_datetime_UTC"]
            )
            if len(locations_for_each_example) > 0:
                locations_for_each_example_of_each_split[split_name] = locations_for_each_example

        # Fire up a separate process for each DataSource, and pass it a list of batches to
        # create, and whether to utils.upload_and_delete_local_files().
        # TODO: Issue 321: Split this up into separate functions!!!
        n_data_sources = len(self.data_sources)
        nd_utils.set_fsspec_for_multiprocess()
        for split_name, locations_for_split in locations_for_each_example_of_each_split.items():
            with multiprocessing.Pool(processes=n_data_sources) as pool:
                async_results_from_create_batches = []
                an_error_has_occured = multiprocessing.Event()
                for worker_id, (data_source_name, data_source) in enumerate(
                    self.data_sources.items()
                ):

                    # Get indexes of first batch and example. And subset locations_for_split.
                    idx_of_first_batch = first_batches_to_create[split_name][data_source_name]
                    idx_of_first_example = idx_of_first_batch * self.config.process.batch_size
                    locations = locations_for_split.loc[idx_of_first_example:]

                    # Get paths.
                    dst_path = (
                        self.config.output_data.filepath / split_name.value / data_source_name
                    )

                    # TODO: Issue 455: Guarantee that local temp path is unique and empty.
                    local_temp_path = (
                        self.config.process.local_temp_path
                        / split_name.value
                        / data_source_name
                        / f"worker_{worker_id}"
                    )

                    # Make folders.
                    nd_fs_utils.makedirs(dst_path, exist_ok=True)
                    if self.save_batches_locally_and_upload:
                        nd_fs_utils.makedirs(local_temp_path, exist_ok=True)

                    # Key word arguments to be passed into data_source.create_batches():
                    kwargs_for_create_batches = dict(
                        spatial_and_temporal_locations_of_each_example=locations,
                        idx_of_first_batch=idx_of_first_batch,
                        batch_size=self.config.process.batch_size,
                        dst_path=dst_path,
                        local_temp_path=local_temp_path,
                        upload_every_n_batches=self.config.process.upload_every_n_batches,
                    )

                    # Logger messages for callbacks:
                    def _callback(result):
                        """Create callback for 'pool.apply_async'"""
                        logger.info(
                            f"{data_source_name} has finished created batches for {split_name}!"
                        )

                    def _error_callback(exception, data_source_name):
                        """Create error callback for 'pool.apply_async'

                        Need to pass in data_source_name rather than rely on data_source_name
                        in the outer scope, because otherwise the error message will contain
                        the wrong data_source_name (due to stuff happening concurrently!)
                        """
                        logger.exception(
                            f"Exception raised by {data_source_name} whilst creating batches for"
                            f" {split_name.value}\n{exception.__class__.__name__}: {exception}"
                        )
                        an_error_has_occured.set()

                    # Submit data_source.create_batches task to the worker process.
                    logger.debug(
                        f"About to submit create_batches task for {data_source_name}, {split_name}"
                    )
                    async_result = pool.apply_async(
                        data_source.create_batches,
                        kwds=kwargs_for_create_batches,
                        callback=_callback,
                        error_callback=partial(_error_callback, data_source_name=data_source_name),
                    )
                    async_results_from_create_batches.append(async_result)

                # Wait for all async_results to finish:
                for async_result in async_results_from_create_batches:
                    async_result.wait()
                    if an_error_has_occured.is_set():
                        # An error has occurred but, at this point in the code, we don't know which
                        # worker process raised the exception.  But, with luck, the worker process
                        # will have logged an informative exception via the _error_callback func.
                        raise RuntimeError(
                            f"A worker process raised an exception whilst working on {split_name}!"
                        )

                logger.info(f"Finished creating batches for {split_name}!")



""" Functions to make fake coordinates """

from typing import List, Optional

import numpy as np

from nowcasting_dataset.geospatial import lat_lon_to_osgb


def create_random_point_coordinates_osgb(
    size: int, x_center_osgb: Optional = None, y_center_osgb: Optional = None
):
    """Make random coords [OSGB] for pv site, or gsp"""
    # this is about 100KM
    HUNDRED_KILOMETERS = 10 ** 5
    x = np.random.randint(0, HUNDRED_KILOMETERS, size)
    y = np.random.randint(0, HUNDRED_KILOMETERS, size)

    return add_uk_centroid_osgb(x, y, x_center_osgb=x_center_osgb, y_center_osgb=y_center_osgb)


def make_random_image_coords_osgb(
    size: int, x_center_osgb: Optional = None, y_center_osgb: Optional = None
):
    """Make random coords for image. These are ranges for the pixels"""

    ONE_KILOMETER = 10 ** 3

    # 4 kilometer spacing seemed about right for real satellite images
    x = 4 * ONE_KILOMETER * np.array((range(0, size)))
    y = 4 * ONE_KILOMETER * np.array((range(size, 0, -1)))

    return add_uk_centroid_osgb(
        x, y, x_center_osgb=x_center_osgb, y_center_osgb=y_center_osgb, first_value_center=False
    )


def make_random_x_and_y_osgb_centers(batch_size: int) -> (List[int], List[int]):
    """Make X and Y OSGB centers"""
    lat = np.random.uniform(51, 55, batch_size)
    lon = np.random.uniform(-2.5, 0.5, batch_size)
    x_centers_osgb, y_centers_osgb = lat_lon_to_osgb(lat=lat, lon=lon)

    return x_centers_osgb, y_centers_osgb


def add_uk_centroid_osgb(
    x,
    y,
    x_center_osgb: Optional = None,
    y_center_osgb: Optional = None,
    first_value_center: bool = True,
):
    """
    Add an OSGB value to make in center of UK

    Args:
        x: random values, OSGB
        y: random values, OSGB
        y_center_osgb: TODO
        x_center_osgb: TODO
        first_value_center: TODO

    Returns: X,Y random coordinates [OSGB]
    """

    if (x_center_osgb is None) and (y_center_osgb is None):
        x_centers_osgb, y_centers_osgb = make_random_x_and_y_osgb_centers(batch_size=1)
        x_center_osgb = x_centers_osgb[0]
        y_center_osgb = y_centers_osgb[0]

    # normalize
    x = x - x.mean()
    y = y - y.mean()

    # put in the uk
    x = x + x_center_osgb
    y = y + y_center_osgb

    # make first value
    if first_value_center:
        x[0] = x_center_osgb
        y[0] = y_center_osgb

    return x, y



""" Topological DataSource """
import logging
from dataclasses import dataclass
from numbers import Number

import pandas as pd
import rioxarray
import xarray as xr
from rasterio.warp import Resampling

import nowcasting_dataset.filesystem.utils as nd_fs_utils
from nowcasting_dataset.data_sources.data_source import ImageDataSource
from nowcasting_dataset.data_sources.topographic.topographic_model import Topographic
from nowcasting_dataset.geospatial import OSGB
from nowcasting_dataset.utils import OpenData

logger = logging.getLogger(__name__)


@dataclass
class TopographicDataSource(ImageDataSource):
    """Add topographic/elevation map features."""

    filename: str = None

    def __post_init__(self, image_size_pixels: int, meters_per_pixel: int):
        """Post init"""
        super().__post_init__(image_size_pixels, meters_per_pixel)
        self._shape_of_example = (
            image_size_pixels,
            image_size_pixels,
        )

        logger.info(f"Loading Topological data {self.filename}")

        with OpenData(file_name=self.filename) as filename:
            self._data = rioxarray.open_rasterio(
                filename=filename, parse_coordinates=True, masked=True
            )

        self._data = self._data.fillna(0)  # Set nodata values to 0 (mostly should be ocean)
        # Add CRS for later, topo maps are assumed to be in OSGB
        self._data.attrs["crs"] = OSGB
        # Distance between pixels, giving their spatial extant, in meters
        self._stored_pixel_size_meters = abs(self._data.coords["x"][1] - self._data.coords["x"][0])
        self._meters_per_pixel = meters_per_pixel

    @staticmethod
    def get_data_model_for_batch():
        """Get the model that is used in the batch"""
        return Topographic

    def check_input_paths_exist(self) -> None:
        """Check input paths exist.  If not, raise a FileNotFoundError."""
        nd_fs_utils.check_path_exists(self.filename)

    def get_example(
        self, t0_dt: pd.Timestamp, x_meters_center: Number, y_meters_center: Number
    ) -> xr.Dataset:
        """
        Get a single example

        Args:
            t0_dt: Current datetime for the example, unused
            x_meters_center: Center of the example in meters in the x direction in OSGB coordinates
            y_meters_center: Center of the example in meters in the y direction in OSGB coordinates

        Returns:
            Example containing topographic data for the selected area
        """
        bounding_box = self._square.bounding_box_centered_on(
            x_meters_center=x_meters_center, y_meters_center=y_meters_center
        )
        selected_data = self._data.sel(
            x=slice(bounding_box.left, bounding_box.right),
            y=slice(bounding_box.top, bounding_box.bottom),
        )
        if self._stored_pixel_size_meters != self._meters_per_pixel:
            # Rescale here to the exact size, assumes that the above is good slice
            # Useful if using different spatially sized grids
            selected_data = selected_data.rio.reproject(
                dst_crs=selected_data.attrs["crs"],
                shape=(self._square.size_pixels, self._square.size_pixels),
                resampling=Resampling.bilinear,
            )

        # selected_data is likely to have 1 too many pixels in x and y
        # because sel(x=slice(a, b)) is [a, b], not [a, b).  So trim:
        selected_data = selected_data.isel(
            x=slice(0, self._square.size_pixels), y=slice(0, self._square.size_pixels)
        )

        selected_data = self._post_process_example(selected_data, t0_dt)
        if selected_data.shape != self._shape_of_example:
            raise RuntimeError(
                "Example is wrong shape! "
                f"x_meters_center={x_meters_center}\n"
                f"y_meters_center={y_meters_center}\n"
                f"t0_dt={t0_dt}\n"
                f"expected shape={self._shape_of_example}\n"
                f"actual shape {selected_data.shape}"
            )

        # change to dataset
        topo_xd = selected_data.to_dataset(name="data")

        return topo_xd

    def _post_process_example(
        self, selected_data: xr.DataArray, t0_dt: pd.Timestamp
    ) -> xr.DataArray:
        """
        Post process the topographical data, removing an extra dim

        Args:
            selected_data: DataArray containing the topographic data
            t0_dt: Unused

        Returns:
            DataArray  removed first dimension
        """
        # Shrink extra dims
        selected_data = selected_data.squeeze()
        return selected_data



""" GSP Data Source. GSP - Grid Supply Points

Read more https://data.nationalgrideso.com/system/gis-boundaries-for-gb-grid-supply-points
"""
import logging
from dataclasses import dataclass
from datetime import datetime
from numbers import Number
from pathlib import Path
from typing import List, Optional, Tuple, Union

import numpy as np
import pandas as pd
import xarray as xr

import nowcasting_dataset.filesystem.utils as nd_fs_utils
from nowcasting_dataset.consts import DEFAULT_N_GSP_PER_EXAMPLE
from nowcasting_dataset.data_sources.data_source import ImageDataSource
from nowcasting_dataset.data_sources.gsp.eso import get_gsp_metadata_from_eso
from nowcasting_dataset.data_sources.gsp.gsp_model import GSP
from nowcasting_dataset.geospatial import lat_lon_to_osgb
from nowcasting_dataset.square import get_bounding_box_mask

logger = logging.getLogger(__name__)


@dataclass
class GSPDataSource(ImageDataSource):
    """
    Data source for GSP (Grid Supply Point) PV Data.

    30 mins data is taken from 'PV Live' from https://www.solar.sheffield.ac.uk/pvlive/
    meta data is taken from ESO.  PV Live estimates the total PV power generation for each
    Grid Supply Point region.

    Even though GSP data isn't image data, `GSPDataSource` inherits from `ImageDataSource`
    so it can select Grid Supply Point regions within the geospatial region of interest.
    The region of interest is defined by `image_size_pixels` and `meters_per_pixel`.
    """

    # zarr_path of where the gsp data is stored
    zarr_path: Union[str, Path]
    # start datetime, this can be None
    start_datetime: Optional[datetime] = None
    # end datetime, this can be None
    end_datetime: Optional[datetime] = None
    # the threshold where we only taken gsp's with a maximum power, above this value.
    threshold_mw: int = 0
    # get the data for the gsp at the center too.
    # This can be turned off if the center of the bounding box is of a pv system
    get_center: bool = True
    # the maximum number of gsp's to be loaded for data sample
    n_gsp_per_example: int = DEFAULT_N_GSP_PER_EXAMPLE
    # scale from zero to one
    do_scale_0_to_1: bool = False
    # Drop any GSPs norther of this boundary.
    # Many satellite images have a "rectangle of zeros" extending north from 1,037,047 meters
    # (OSGB "northing" coordinate).  The largest satellite regions of interest are
    # 144 km.  So set the boundary to be 1,037 km - (144 km / 2) = 1,036,975 meters.
    # This results in 2 GSPs being dropped.
    # See https://github.com/openclimatefix/Satip/issues/30
    northern_boundary_osgb: Optional[float] = 1_036_975

    def __post_init__(self, image_size_pixels: int, meters_per_pixel: int):
        """
        Set random seed and load data
        """
        super().__post_init__(image_size_pixels, meters_per_pixel)
        self.rng = np.random.default_rng()
        self.load()

    def check_input_paths_exist(self) -> None:
        """Check input paths exist.  If not, raise a FileNotFoundError."""
        nd_fs_utils.check_path_exists(self.zarr_path)

    @property
    def sample_period_minutes(self) -> int:
        """Override the default sample minutes"""
        return 30

    @staticmethod
    def get_data_model_for_batch():
        """Get the model that is used in the batch"""
        return GSP

    def load(self):
        """
        Load the meta data and load the GSP power data
        """
        # load metadata
        self.metadata = get_gsp_metadata_from_eso()
        self.metadata.set_index("gsp_id", drop=False, inplace=True)

        # make location x,y in osgb
        self.metadata["location_x"], self.metadata["location_y"] = lat_lon_to_osgb(
            lat=self.metadata["centroid_lat"], lon=self.metadata["centroid_lon"]
        )

        # load gsp data from file / gcp
        self.gsp_power, self.gsp_capacity = load_solar_gsp_data(
            self.zarr_path, start_dt=self.start_datetime, end_dt=self.end_datetime
        )

        # drop any gsp below a threshold mw. This is to get rid of any small GSP where
        # predicting the solar output will be harder.
        self.gsp_power, self.metadata = drop_gsp_by_threshold(
            self.gsp_power, self.metadata, threshold_mw=self.threshold_mw
        )

        if self.northern_boundary_osgb is not None:
            self.gsp_power, self.metadata = drop_gsp_north_of_boundary(
                self.gsp_power, self.metadata, northern_boundary_osgb=self.northern_boundary_osgb
            )

        logger.debug(f"There are {len(self.gsp_power.columns)} GSP")

    def datetime_index(self):
        """
        Return the datetimes that are available
        """
        return self.gsp_power.index

    def get_number_locations(self):
        """Get the number of GSP"""

        return len(self.metadata.location_x)

    def get_all_locations(
        self, t0_datetimes_utc: pd.DatetimeIndex
    ) -> Tuple[pd.DatetimeIndex, List[Number], List[Number]]:
        """
        Make locations for all GSP

        For some datetimes, return locations of all datetimes and all GSPs.
        This means a national forecast can then be made

        Args:
            t0_datetimes_utc: list of available t0 datetimes.

        Returns:
            1. list of datetimes
            2. list of x locations
            3. list of y locations

        """

        logger.info("Getting all locations for each datetime")

        total_gsp_nan_count = self.gsp_power.isna().sum().sum()
        if total_gsp_nan_count > 0:
            assert Exception("There are nans in the GSP data. Can't get locations for all GSPs")
        else:

            t0_datetimes_utc.name = "t0_datetime_utc"

            # get all locations
            x_centers_osgb = self.metadata.location_x
            y_centers_osgb = self.metadata.location_y

            # make x centers
            x_centers_osgb_all_gsps = pd.DataFrame(columns=t0_datetimes_utc, index=x_centers_osgb)
            x_centers_osgb_all_gsps = x_centers_osgb_all_gsps.unstack().reset_index()

            # make y centers
            y_centers_osgb_all_gsps = pd.DataFrame(columns=t0_datetimes_utc, index=y_centers_osgb)
            y_centers_osgb_all_gsps = y_centers_osgb_all_gsps.unstack().reset_index()

            t0_datetimes_utc_all_gsps = pd.DatetimeIndex(x_centers_osgb_all_gsps["t0_datetime_utc"])
            x_centers_osgb_all_gsps = list(x_centers_osgb_all_gsps["location_x"])
            y_centers_osgb_all_gsps = list(y_centers_osgb_all_gsps["location_y"])

            return t0_datetimes_utc_all_gsps, x_centers_osgb_all_gsps, y_centers_osgb_all_gsps

    def get_locations(self, t0_datetimes: pd.DatetimeIndex) -> Tuple[List[Number], List[Number]]:
        """
        Get x and y locations. Assume that all data is available for all GSP.

        Random GSP are taken, and the locations of them are returned. This is useful as other
        datasources need to know which x,y locations to get.

        Args:
            t0_datetimes: list of available t0 datetimes.

        Returns: list of x and y locations

        """

        logger.info("Getting one location for each datetime")

        total_gsp_nan_count = self.gsp_power.isna().sum().sum()
        if total_gsp_nan_count == 0:

            # get random GSP metadata
            indexes = list(
                self.rng.integers(low=0, high=len(self.metadata), size=len(t0_datetimes))
            )
            metadata = self.metadata.iloc[indexes]

            # get x, y locations
            x_centers_osgb = list(metadata.location_x)
            y_centers_osgb = list(metadata.location_y)

        else:

            logger.warning(
                "There are some nans in the gsp data, "
                "so to get x,y locations we have to do a big loop"
            )

            # Pick a random GSP for each t0_datetime, and then grab
            # their geographical location.
            x_centers_osgb = []
            y_centers_osgb = []

            for t0_dt in t0_datetimes:

                # Choose start and end times
                start_dt = self._get_start_dt(t0_dt)
                end_dt = self._get_end_dt(t0_dt)

                # remove any nans
                gsp_power = self.gsp_power.loc[start_dt:end_dt].dropna(axis="columns", how="any")

                # get random index
                random_gsp_id = self.rng.choice(gsp_power.columns)
                meta_data = self.metadata[(self.metadata["gsp_id"] == random_gsp_id)]

                # Make sure there is only one GSP.
                # Sometimes there are multiple gsp_ids at one location e.g. 'SELL_1'.
                # TODO: Issue #272: Further investigation on multiple GSPs may be needed.
                metadata_for_gsp = meta_data.iloc[0]

                # Get metadata for GSP
                x_centers_osgb.append(metadata_for_gsp.location_x)
                y_centers_osgb.append(metadata_for_gsp.location_y)

        return x_centers_osgb, y_centers_osgb

    def get_example(
        self, t0_dt: pd.Timestamp, x_meters_center: Number, y_meters_center: Number
    ) -> xr.Dataset:
        """
        Get data example from one time point (t0_dt) and for x and y coords.

        Get data at the location of x,y and get surrounding GSP power data also.

        Args:
            t0_dt: datetime of "now". History and forecast are also returned
            x_meters_center: x location of center GSP.
            y_meters_center: y location of center GSP.

        Returns: Dictionary with GSP data in it.
        """
        logger.debug("Getting example data")

        # get the GSP power, including history and forecast
        selected_gsp_power, selected_capacity = self._get_time_slice(t0_dt)

        # get the main gsp id, and the ids of the gsp in the bounding box
        all_gsp_ids = self._get_gsp_ids_in_roi(
            x_meters_center, y_meters_center, selected_gsp_power.columns
        )
        if self.get_center:
            central_gsp_id = self._get_central_gsp_id(
                x_meters_center, y_meters_center, selected_gsp_power.columns
            )
            assert central_gsp_id in all_gsp_ids

            # By convention, the 'target' GSP ID (the one in the center
            # of the image) must be in the first position of the returned arrays.
            all_gsp_ids = all_gsp_ids.drop(central_gsp_id)
            all_gsp_ids = all_gsp_ids.insert(loc=0, item=central_gsp_id)
        else:
            logger.warning("Not getting center GSP")

        # only select at most {n_gsp_per_example}
        all_gsp_ids = all_gsp_ids[: self.n_gsp_per_example]

        # select the GSP power output for the selected GSP IDs
        selected_gsp_power = selected_gsp_power[all_gsp_ids]
        selected_capacity = selected_capacity[all_gsp_ids]

        # get x,y coordinates
        gsp_x_coords = self.metadata.location_x[all_gsp_ids]
        gsp_y_coords = self.metadata.location_y[all_gsp_ids]

        # convert to data array
        da = xr.DataArray(
            data=selected_gsp_power.values,
            dims=["time", "id"],
            coords=dict(
                id=all_gsp_ids.values.astype(int),
                time=selected_gsp_power.index.values,
            ),
        )

        capacity = xr.DataArray(
            data=selected_capacity.values,
            dims=["time", "id"],
            coords=dict(
                id=all_gsp_ids.values.astype(int),
                time=selected_gsp_power.index.values,
            ),
        )

        # convert to dataset
        gsp = da.to_dataset(name="power_mw")
        gsp["capacity_mwp"] = capacity

        # add gsp x coords
        gsp_x_coords = xr.DataArray(
            data=gsp_x_coords.values,
            dims=["id"],
        )

        gsp_y_coords = xr.DataArray(
            data=gsp_y_coords.values,
            dims=["id"],
        )
        gsp["x_coords"] = gsp_x_coords
        gsp["y_coords"] = gsp_y_coords

        # pad out so that there are always 32 gsp, fill with 0
        pad_n = self.n_gsp_per_example - len(gsp.id)
        gsp = gsp.pad(id=(0, pad_n), power_mw=((0, 0), (0, pad_n)), constant_values=0)

        return gsp

    def _get_central_gsp_id(
        self,
        x_meters_center: Number,
        y_meters_center: Number,
        gsp_ids_with_data_for_timeslice: pd.Int64Index,
    ) -> int:
        """
        Get the GSP id of the central GSP from coordinates

        Args:
            x_meters_center: the location of the gsp (x)
            y_meters_center: the location of the gsp (y)
            gsp_ids_with_data_for_timeslice: List of gsp ids that are available for a certain
                timeslice.

        Returns: GSP id
        """
        logger.debug("Getting Central GSP")

        # If x_meters_center and y_meters_center have been chosen
        # by {}.get_locations() then we just have
        # to find the gsp_ids at that exact location.  This is
        # super-fast (a few hundred microseconds).  We use np.isclose
        # instead of the equality operator because floats.
        meta_data_index = self.metadata.index[
            np.isclose(self.metadata.location_x, x_meters_center, rtol=1e-05, atol=1e-05)
            & np.isclose(self.metadata.location_y, y_meters_center, rtol=1e-05, atol=1e-05)
        ]
        gsp_ids = self.metadata.loc[meta_data_index].gsp_id.values

        if len(gsp_ids) == 0:
            # TODO: Implement finding GSP closest to x_meters_center,
            # y_meters_center.  This will probably be quite slow, so always
            # try finding an exact match first (which is super-fast).
            raise NotImplementedError(
                "Not yet implemented the ability to find GSP *nearest*"
                " (but not at the identical location to) x_meters_center and"
                " y_meters_center."
            )

        gsp_ids = gsp_ids_with_data_for_timeslice.intersection(gsp_ids)

        if len(gsp_ids) == 0:
            raise NotImplementedError(
                f"Could not find GSP id for {x_meters_center}, {y_meters_center} "
                f"({gsp_ids}) and {gsp_ids_with_data_for_timeslice}"
            )

        return int(gsp_ids[0])

    def _get_gsp_ids_in_roi(
        self,
        x_meters_center: Number,
        y_meters_center: Number,
        gsp_ids_with_data_for_timeslice: pd.Int64Index,
    ) -> pd.Int64Index:
        """
        Find the GSP IDs for all the GSP within the geospatial region of interest.

        The geospatial region of interest is defined by self.square.

        Args:
            x_meters_center: center of area of interest (x coords)
            y_meters_center: center of area of interest (y coords)
            gsp_ids_with_data_for_timeslice: ids that are avialble for a specific time slice

        Returns: list of GSP ids that are in area of interest
        """
        logger.debug(f"Getting all gsp in ROI ({x_meters_center=},{y_meters_center=})")

        # creating bounding box
        bounding_box = self._square.bounding_box_centered_on(
            x_meters_center=x_meters_center, y_meters_center=y_meters_center
        )

        # get all x and y locations of gsp
        x = self.metadata.location_x
        y = self.metadata.location_y

        # make mask of gsp_ids
        mask = get_bounding_box_mask(bounding_box, x, y)

        gsp_ids = self.metadata[mask].gsp_id
        gsp_ids = gsp_ids_with_data_for_timeslice.intersection(gsp_ids)

        assert len(gsp_ids) > 0
        return gsp_ids

    def _get_time_slice(self, t0_dt: pd.Timestamp) -> [pd.DataFrame]:
        """
        Get time slice of GSP power data for give time.

        Note the time is extended backwards by history lenght and forward by prediction time

        Args:
            t0_dt: timestamp of interest

        Returns: pandas data frame of GSP power data
        """
        logger.debug(f"Getting power slice for {t0_dt}")

        # get start and end datetime, takening into account history and forecast length.
        start_dt = self._get_start_dt(t0_dt)
        end_dt = self._get_end_dt(t0_dt)

        # need to floor by 30 mins.
        # If t0 is 12.45 and history duration is 1 hours, then start_dt will be 11.45.
        # But we need to collect data at 11.30, 12.00, and 12.30
        start_dt = pd.to_datetime(start_dt).floor("30T")

        # select power and capacity for certain times
        power = self.gsp_power.loc[start_dt:end_dt]
        capacity = self.gsp_capacity.loc[start_dt:end_dt]

        # remove any nans
        power = power.dropna(axis="columns", how="any")
        capacity = capacity.dropna(axis="columns", how="any")

        logger.debug(f"Found {len(power.columns)} GSP valid data for {t0_dt}")

        return power, capacity


def drop_gsp_by_threshold(
    gsp_power: pd.DataFrame, meta_data: pd.DataFrame, threshold_mw: int = 20
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """
    Drop GSP where the max power is below a certain threshold

    Args:
        gsp_power: GSP power data
        meta_data: the GSP meta data
        threshold_mw: the threshold where we only taken GSP with a maximum power, above this value.

    Returns: power data and metadata
    """
    maximum_gsp = gsp_power.max()

    keep_index = maximum_gsp > threshold_mw

    dropped_gsp_names = meta_data.loc[keep_index[~keep_index].index].gsp_name

    logger.debug(
        f"Dropping {sum(~keep_index)} GSPs as maximum is not greater {threshold_mw} MW."
        f" Dropped GSP IDs and GSP names:\n{dropped_gsp_names}"
    )
    logger.debug(f"Keeping {sum(keep_index)} GSPs as maximum is greater {threshold_mw} MW")

    filtered_gsp_power = gsp_power.loc[:, keep_index]
    filtered_gsp_ids = filtered_gsp_power.columns
    filtered_meta_data = meta_data[meta_data["gsp_id"].isin(filtered_gsp_ids)]

    assert set(filtered_gsp_power.columns) == set(filtered_meta_data.gsp_id)
    return filtered_gsp_power, filtered_meta_data


def drop_gsp_north_of_boundary(
    gsp_power: pd.DataFrame, meta_data: pd.DataFrame, northern_boundary_osgb: int
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """
    Drop GSPs north of northern_boundary_osgb.

    Args:
        gsp_power: GSP power data
        meta_data: the GSP meta data
        northern_boundary_osgb: The geospatial boundary.

    Returns: power data and metadata
    """

    keep_index = meta_data.location_y < northern_boundary_osgb
    filtered_meta_data = meta_data.loc[keep_index]
    filtered_gsp_ids = filtered_meta_data.gsp_id
    filtered_gsp_power = gsp_power[filtered_gsp_ids]
    pv_capacity_of_dropped_gsps = gsp_power.loc[:, ~keep_index].max()

    logger.debug(
        f"Dropping {sum(~keep_index)} GSPs because they are north of"
        f" y={northern_boundary_osgb:,d}m:\n"
        f" {meta_data.loc[~keep_index].gsp_name}\n with capacity in MW of:\n"
        f"{pv_capacity_of_dropped_gsps}"
    )
    logger.debug(
        f"Keeping {sum(keep_index)} GSPs because they are south of y={northern_boundary_osgb:,d}m."
    )

    assert set(filtered_gsp_power.columns) == set(filtered_meta_data.gsp_id)
    return filtered_gsp_power, filtered_meta_data


def load_solar_gsp_data(
    zarr_path: Union[str, Path],
    start_dt: Optional[datetime] = None,
    end_dt: Optional[datetime] = None,
) -> (pd.DataFrame, pd.DataFrame):
    """
    Load solar PV GSP data

    Args:
        zarr_path:  zarr_path of file to be loaded, can put 'gs://' files in here too
        start_dt: the start datetime, which to trim the data to
        end_dt: the end datetime, which to trim the data to

    Returns: dataframe of pv data

    """
    logger.debug(f"Loading Solar GSP Data from {zarr_path} from {start_dt} to {end_dt}")
    # Open data - it may be quicker to open byte file first, but decided just to keep it
    # like this at the moment.
    gsp_power_and_capacity = xr.open_dataset(zarr_path, engine="zarr")
    gsp_power_and_capacity = gsp_power_and_capacity.sel(datetime_gmt=slice(start_dt, end_dt))

    # make dataframe with index datetime_gmt and columns og gsp_id
    gsp_power_and_capacity_df = gsp_power_and_capacity.to_dataframe()
    gsp_power_and_capacity_df.reset_index(inplace=True)
    gsp_power_df = gsp_power_and_capacity_df.pivot(
        index="datetime_gmt", columns="gsp_id", values="generation_mw"
    )

    gsp_capacity_df = gsp_power_and_capacity_df.pivot(
        index="datetime_gmt", columns="gsp_id", values="installedcapacity_mwp"
    )

    # Save memory
    del gsp_power_and_capacity

    # Process the data a little
    gsp_power_df = gsp_power_df.dropna(axis="columns", how="all")
    gsp_power_df = gsp_power_df.clip(lower=0, upper=5e7)
    gsp_capacity_df = gsp_capacity_df.dropna(axis="columns", how="all")
    gsp_capacity_df = gsp_capacity_df.clip(lower=0, upper=5e7)

    # remove whole rows of nans
    gsp_power_df = gsp_power_df.dropna(axis="columns", how="all")
    gsp_capacity_df = gsp_capacity_df.loc[:, gsp_power_df.columns]

    # make column names ints, not strings
    gsp_power_df.columns = [int(col) for col in gsp_power_df.columns]
    gsp_capacity_df.columns = [int(col) for col in gsp_capacity_df.columns]

    return gsp_power_df, gsp_capacity_df



""" Satellite Data Source """
import logging
from dataclasses import InitVar, dataclass
from functools import partial
from numbers import Number
from typing import Iterable, Optional

import dask
import numpy as np
import pandas as pd
import xarray as xr

import nowcasting_dataset.time as nd_time
from nowcasting_dataset.consts import SAT_VARIABLE_NAMES
from nowcasting_dataset.data_sources.data_source import ZarrDataSource
from nowcasting_dataset.data_sources.satellite.satellite_model import Satellite

_LOG = logging.getLogger(__name__)
_LOG_HRV = logging.getLogger(__name__.replace("satellite", "hrvsatellite"))


@dataclass
class SatelliteDataSource(ZarrDataSource):
    """Satellite Data Source."""

    channels: Optional[Iterable[str]] = SAT_VARIABLE_NAMES[1:]
    image_size_pixels: InitVar[int] = 128
    meters_per_pixel: InitVar[int] = 2_000
    logger = _LOG

    def __post_init__(self, image_size_pixels: int, meters_per_pixel: int):
        """Post Init"""
        assert len(self.channels) > 0, "channels cannot be empty!"
        assert image_size_pixels > 0, "image_size_pixels cannot be <= 0!"
        assert meters_per_pixel > 0, "meters_per_pixel cannot be <= 0!"
        super().__post_init__(image_size_pixels, meters_per_pixel)
        n_channels = len(self.channels)
        self._shape_of_example = (
            self.total_seq_length,
            image_size_pixels,
            image_size_pixels,
            n_channels,
        )

    def open(self) -> None:
        """
        Open Satellite data

        We don't want to open_sat_data in __init__.
        If we did that, then we couldn't copy SatelliteDataSource
        instances into separate processes.  Instead,
        call open() _after_ creating separate processes.
        """
        self._data = self._open_data()
        if "variable" in self._data.dims:
            self._data = self._data.rename({"variable": "channels"})
        if not set(self.channels).issubset(self._data.channels.values):
            raise RuntimeError(
                f"One or more requested channels are not available in {self.zarr_path}!"
                f"  Requested channels={self.channels}."
                f"  Available channels={self._data.channels.values}"
            )
        self._data = self._data.sel(channels=list(self.channels))

    def _open_data(self) -> xr.DataArray:
        return open_sat_data(
            zarr_path=self.zarr_path, consolidated=self.consolidated, logger=self.logger
        )

    @staticmethod
    def get_data_model_for_batch():
        """Get the model that is used in the batch"""
        return Satellite

    def _get_time_slice(self, t0_dt: pd.Timestamp) -> xr.DataArray:
        start_dt = self._get_start_dt(t0_dt)
        end_dt = self._get_end_dt(t0_dt)
        data = self.data.sel(time=slice(start_dt, end_dt))
        assert type(data) == xr.DataArray

        return data

    def get_spatial_region_of_interest(
        self, data_array: xr.DataArray, x_center_osgb: Number, y_center_osgb: Number
    ) -> xr.DataArray:
        """
        Gets the satellite image as a square around the center

        Ignores x and y coordinates as for the original satellite projection each pixel varies in
        both its x and y distance from other pixels. See Issue 401 for more details.

        This results, in 'real' spatial terms, each image covering about 2x as much distance in the
        x direction as in the y direction.

        Args:
            data_array: DataArray to subselect from
            x_center_osgb: Center of the image x coordinate in OSGB coordinates
            y_center_osgb: Center of image y coordinate in OSGB coordinates

        Returns:
            The selected data around the center
        """
        # Get the index into x and y nearest to x_center_osgb and y_center_osgb:
        x_index_at_center = np.searchsorted(data_array.x.values, x_center_osgb) - 1
        y_index_at_center = np.searchsorted(data_array.y.values, y_center_osgb) - 1
        # Put x_index_at_center and y_index_at_center into a pd.Series so we can operate
        # on them both in a single line of code.
        x_and_y_index_at_center = pd.Series({"x": x_index_at_center, "y": y_index_at_center})
        half_image_size_pixels = self._square.size_pixels // 2
        min_x_and_y_index = x_and_y_index_at_center - half_image_size_pixels
        max_x_and_y_index = x_and_y_index_at_center + half_image_size_pixels

        # Check whether the requested region of interest steps outside of the available data:
        suggested_reduction_of_image_size_pixels = (
            max(
                (-min_x_and_y_index.min() if (min_x_and_y_index < 0).any() else 0),
                (max_x_and_y_index.x - len(data_array.x)),
                (max_x_and_y_index.y - len(data_array.y)),
            )
            * 2
        )
        # If the requested region does step outside the available data then raise an exception
        # with a helpful message:
        if suggested_reduction_of_image_size_pixels > 0:
            new_suggested_image_size_pixels = (
                self._square.size_pixels - suggested_reduction_of_image_size_pixels
            )
            raise RuntimeError(
                "Requested region of interest of satellite data steps outside of the available"
                " geographical extent of the Zarr data.  The requested region of interest extends"
                f" from pixel indicies"
                f" x={min_x_and_y_index.x} to x={max_x_and_y_index.x},"
                f" y={min_x_and_y_index.y} to y={max_x_and_y_index.y}.  In the Zarr data,"
                f" len(x)={len(data_array.x)}, len(y)={len(data_array.y)}. Try reducing"
                f" image_size_pixels from {self._square.size_pixels} to"
                f" {new_suggested_image_size_pixels} pixels."
            )

        # Select the geographical region of interest.
        # Note that isel is *exclusive* of the end of the slice.
        # e.g. isel(x=slice(0, 3)) will return the first, second, and third values.
        data_array = data_array.isel(
            x=slice(min_x_and_y_index.x, max_x_and_y_index.x),
            y=slice(min_x_and_y_index.y, max_x_and_y_index.y),
        )
        return data_array

    def get_example(
        self, t0_dt: pd.Timestamp, x_meters_center: Number, y_meters_center: Number
    ) -> xr.Dataset:
        """
        Get Example data

        Args:
            t0_dt: list of timestamps for the datetime of the batches. The batch will also include
                data for historic and future depending on `history_minutes` and `future_minutes`.
            x_meters_center: x center batch locations
            y_meters_center: y center batch locations

        Returns: Example Data

        """
        selected_data = self._get_time_slice(t0_dt)
        selected_data = self.get_spatial_region_of_interest(
            data_array=selected_data,
            x_center_osgb=x_meters_center,
            y_center_osgb=y_meters_center,
        )

        if "variable" in list(selected_data.dims):
            selected_data = selected_data.rename({"variable": "channels"})

        selected_data = self._post_process_example(selected_data, t0_dt)

        if selected_data.shape != self._shape_of_example:
            raise RuntimeError(
                "Example is wrong shape! "
                f"x_meters_center={x_meters_center}\n"
                f"y_meters_center={y_meters_center}\n"
                f"t0_dt={t0_dt}\n"
                f"times are {selected_data.time}\n"
                f"expected shape={self._shape_of_example}\n"
                f"actual shape {selected_data.shape}"
            )

        return selected_data.load().to_dataset(name="data")

    def datetime_index(self, remove_night: bool = True) -> pd.DatetimeIndex:
        """Returns a complete list of all available datetimes

        Args:
            remove_night: If True then remove datetimes at night.
                We're interested in forecasting solar power generation, so we
                don't care about nighttime data :)

                In the UK in summer, the sun rises first in the north east, and
                sets last in the north west [1].  In summer, the north gets more
                hours of sunshine per day.

                In the UK in winter, the sun rises first in the south east, and
                sets last in the south west [2].  In winter, the south gets more
                hours of sunshine per day.

                |                        | Summer | Winter |
                |           ---:         |  :---: |  :---: |
                | Sun rises first in     | N.E.   | S.E.   |
                | Sun sets last in       | N.W.   | S.W.   |
                | Most hours of sunlight | North  | South  |

                Before training, we select timesteps which have at least some
                sunlight.  We do this by computing the clearsky global horizontal
                irradiance (GHI) for the four corners of the satellite imagery,
                and for all the timesteps in the dataset.  We only use timesteps
                where the maximum global horizontal irradiance across all four
                corners is above some threshold.

                The 'clearsky solar irradiance' is the amount of sunlight we'd
                expect on a clear day at a specific time and location. The SI unit
                of irradiance is watt per square meter.  The 'global horizontal
                irradiance' (GHI) is the total sunlight that would hit a
                horizontal surface on the surface of the Earth.  The GHI is the
                sum of the direct irradiance (sunlight which takes a direct path
                from the Sun to the Earth's surface) and the diffuse horizontal
                irradiance (the sunlight scattered from the atmosphere).  For more
                info, see: https://en.wikipedia.org/wiki/Solar_irradiance

        References:
          1. [Video of June 2019](https://www.youtube.com/watch?v=IOp-tj-IJpk)
          2. [Video of Jan 2019](https://www.youtube.com/watch?v=CJ4prUVa2nQ)
        """
        if self._data is None:
            sat_data = self._open_data()
        else:
            sat_data = self._data

        datetime_index = pd.DatetimeIndex(sat_data.time.values)

        if remove_night:
            border_locations = self.geospatial_border()
            datetime_index = nd_time.select_daylight_datetimes(
                datetimes=datetime_index, locations=border_locations
            )

        return datetime_index


class HRVSatelliteDataSource(SatelliteDataSource):
    """Satellite Data Source for HRV data."""

    channels: Optional[Iterable[str]] = SAT_VARIABLE_NAMES[:1]
    image_size_pixels: InitVar[int] = 128
    meters_per_pixel: InitVar[int] = 2_000
    logger = _LOG_HRV


def remove_acq_time_from_dataset_and_fix_time_coords(
    dataset: xr.Dataset, logger: logging.Logger
) -> xr.Dataset:
    """
    Preprocess datasets by dropping `acq_time`, which causes problems otherwise

    Args:
        dataset: xr.Dataset to preprocess
        logger: logger object to write to

    Returns:
        dataset with acq_time dropped
    """
    dataset = dataset.drop_vars("acq_time", errors="ignore")

    # If there are any duplicated init_times then drop the duplicated init_times:
    data_array = dataset["stacked_eumetsat_data"]
    times = pd.DatetimeIndex(data_array["time"])
    if not times.is_unique:
        n_duplicates = times.duplicated().sum()
        logger.warning(f"Satellite Zarr has {n_duplicates:,d} duplicated times.  Fixing...")
        data_array = data_array.drop_duplicates(dim="time")
        times = pd.DatetimeIndex(data_array["time"])
        dataset = data_array.to_dataset(name="stacked_eumetsat_data")

    # If any init_times are not monotonic_increasing then drop the out-of-order init_times:
    if not times.is_monotonic_increasing:
        total_n_out_of_order_times = 0
        logger.warning("Satellite Zarr times is not monotonic_increasing.  Fixing...")
        while not times.is_monotonic_increasing:
            diff = np.diff(times.view(int))
            out_of_order = np.where(diff < 0)[0]
            total_n_out_of_order_times += len(out_of_order)
            out_of_order = times[out_of_order]
            data_array = data_array.drop_sel(time=out_of_order)
            times = pd.DatetimeIndex(data_array["init_time"])
        logger.info(f"Fixed {total_n_out_of_order_times:,d} out of order init_times.")
        dataset = data_array.to_dataset(name="stacked_eumetsat_data")

    return dataset


def open_sat_data(zarr_path: str, consolidated: bool, logger: logging.Logger) -> xr.DataArray:
    """Lazily opens the Zarr store.

    Adds 1 minute to the 'time' coordinates, so the timestamps
    are at 00, 05, ..., 55 past the hour.

    Args:
      zarr_path: Cloud URL or local path pattern.  If GCP URL, must start with 'gs://'
      consolidated: Whether or not the Zarr metadata is consolidated.
      logger: logger object to write to
    """
    logger.debug("Opening satellite data: %s", zarr_path)

    # If we are opening multiple Zarr stores (i.e. one for each month of the year) we load them
    # together and create a single dataset from them.  open_mfdataset also works if zarr_path
    # points to a specific zarr directory (with no wildcards).

    # Silence the warning about large chunks.
    # Alternatively, we could set this to True, but that slows down loading a Satellite batch
    # from 8 seconds to 50 seconds!
    dask.config.set(**{"array.slicing.split_large_chunks": False})

    # add logger to preprocess function
    p_remove_acq_time_from_dataset_and_fix_time_coords = partial(
        remove_acq_time_from_dataset_and_fix_time_coords, logger=logger
    )

    # Open datasets.
    dataset = xr.open_mfdataset(
        zarr_path,
        chunks="auto",  # See issue #456 for why we use "auto".
        mode="r",
        engine="zarr",
        concat_dim="time",
        preprocess=p_remove_acq_time_from_dataset_and_fix_time_coords,
        consolidated=consolidated,
        combine="nested",
    )

    data_array = dataset["stacked_eumetsat_data"]
    if "stacked_eumetsat_data" == data_array.name:
        data_array.name = "data"
    del dataset

    # Flip coordinates to top-left first
    data_array = data_array.reindex(x=data_array.x[::-1])

    # Sanity check!
    times = pd.DatetimeIndex(data_array["time"])
    assert times.is_unique
    assert times.is_monotonic_increasing

    return data_array



""" Useful functions for xarray objects

1. joining data arrays to datasets
2. pydantic exentions model of xr.Dataset
"""
from typing import Any, List

import numpy as np
import xarray as xr


def join_list_dataset_to_batch_dataset(datasets: list[xr.Dataset]) -> xr.Dataset:
    """Join a list of data sets to a dataset by expanding dims"""

    new_datasets = []
    for i, dataset in enumerate(datasets):
        new_dataset = dataset.expand_dims(dim="example").assign_coords(example=("example", [i]))
        new_datasets.append(new_dataset)

    return xr.concat(new_datasets, dim="example")


def convert_coordinates_to_indexes_for_list_datasets(
    examples: List[xr.Dataset],
) -> List[xr.Dataset]:
    """Set the coords to be indices before joining into a batch"""
    return [convert_coordinates_to_indexes(example) for example in examples]


def convert_coordinates_to_indexes(dataset: xr.Dataset) -> xr.Dataset:
    """Reindex dims so that it can be merged with batch.

    For each dimension in dataset, change the coords to 0.. len(original_coords),
    and append "_index" to the dimension name.
    And save the original coordinates in `original_dim_name`.

    This is useful to align multiple examples into a single batch.
    """

    assert type(dataset) == xr.Dataset, f" Should be xr.Dataset but found {type(dataset)}"

    original_dim_names = dataset.dims

    for original_dim_name in original_dim_names:
        original_coords = dataset[original_dim_name]
        new_index_coords = np.arange(len(original_coords))
        new_index_dim_name = f"{original_dim_name}_index"
        dataset[original_dim_name] = new_index_coords
        dataset = dataset.rename({original_dim_name: new_index_dim_name})
        # Save the original_coords back into dataset, but this time it won't be used as
        # coords for the variables payload in the dataset.
        dataset[original_dim_name] = xr.DataArray(
            original_coords,
            coords=[new_index_coords],
            dims=[new_index_dim_name],
        )

    return dataset


class PydanticXArrayDataSet(xr.Dataset):
    """Pydantic Xarray Dataset Class

    Adapted from https://pydantic-docs.helpmanual.io/usage/types/#classes-with-__get_validators__

    """

    _expected_dimensions = ()  # Subclasses should set this.
    _expected_data_vars = ()  # Subclasses should set this.

    # xarray doesnt support sub classing at the moment: https://github.com/pydata/xarray/issues/3980
    __slots__ = ()

    @classmethod
    def model_validation(cls, v):
        """Specific model validation, to be overwritten by class"""
        return v

    @classmethod
    def __get_validators__(cls):
        """Get validators"""
        yield cls.validate

    @classmethod
    def validate(cls, v: Any) -> Any:
        """Do validation"""
        v = cls.validate_dims(v)
        v = cls.validate_coords(v)
        v = cls.validate_data_vars(v)
        v = cls.model_validation(v)
        return v

    @classmethod
    def validate_dims(cls, v: Any) -> Any:
        """Validate the dims"""
        assert all(
            dim.replace("_index", "") in cls._expected_dimensions
            for dim in v.dims
            if dim != "example"
        ), (
            f"{cls.__name__}.dims is wrong! "
            f"{cls.__name__}.dims is {v.dims}. "
            f"But we expected {cls._expected_dimensions}."
            " Note that '_index' is removed, and 'example' is ignored"
        )
        return v

    @classmethod
    def validate_coords(cls, v: Any) -> Any:
        """Validate the coords"""
        for dim in cls._expected_dimensions:
            coord = v.coords[f"{dim}_index"]
            assert len(coord) > 0, f"{dim}_index is empty in {cls.__name__}!"
        return v

    @classmethod
    def validate_data_vars(cls, v: Any) -> Any:
        """Validate the data vars"""

        data_var_names = v.data_vars
        for data_var in cls._expected_data_vars:
            assert (
                data_var in data_var_names
            ), f"{data_var} is not in all data_vars ({data_var_names}) in {cls.__name__}!"
        return v


def convert_arrays_to_uint8(*arrays: tuple[np.ndarray]) -> tuple[np.ndarray]:
    """Convert multiple arrays to uint8, using the same min and max to scale all arrays."""
    # First, stack into a single numpy array so we can work on all images at the same time:
    stacked = np.stack(arrays)

    # Convert to float64 for normalisation:
    stacked = stacked.astype(np.float64)

    # Rescale pixel values to be in the range [0, 1]:
    stacked -= stacked.min()
    stacked_max = stacked.max()
    if stacked_max > 0.0:
        # If there is still an invalid value then we want to know about it!
        # Adapted from https://stackoverflow.com/a/33701974/732596
        with np.errstate(all="raise"):
            stacked /= stacked.max()

    # Convert to uint8 (uint8 can represent integers in the range [0, 255]):
    stacked *= 255
    stacked = stacked.round()
    stacked = stacked.astype(np.uint8)

    return tuple(stacked)



""" Various DataSources """
from nowcasting_dataset.data_sources.data_source import DataSource  # noqa: F401
from nowcasting_dataset.data_sources.gsp.gsp_data_source import GSPDataSource
from nowcasting_dataset.data_sources.nwp.nwp_data_source import NWPDataSource

# We must import OpticalFlowDataSource *after* SatelliteDataSource,
# otherwise we get circular import errors!
from nowcasting_dataset.data_sources.optical_flow.optical_flow_data_source import (
    OpticalFlowDataSource,
)
from nowcasting_dataset.data_sources.pv.pv_data_source import PVDataSource
from nowcasting_dataset.data_sources.satellite.satellite_data_source import (
    HRVSatelliteDataSource,
    SatelliteDataSource,
)
from nowcasting_dataset.data_sources.sun.sun_data_source import SunDataSource
from nowcasting_dataset.data_sources.topographic.topographic_data_source import (
    TopographicDataSource,
)

# If you update MAP_DATA_SOURCE_NAME_TO_CLASS, then please also update ALL_DATA_SOURCE_NAMES in
# nowcasting_dataset.config.model.InputData.set_forecast_and_history_minutes
MAP_DATA_SOURCE_NAME_TO_CLASS = {
    "pv": PVDataSource,
    "satellite": SatelliteDataSource,
    "hrvsatellite": HRVSatelliteDataSource,
    "opticalflow": OpticalFlowDataSource,
    "nwp": NWPDataSource,
    "gsp": GSPDataSource,
    "topographic": TopographicDataSource,
    "sun": SunDataSource,
}
ALL_DATA_SOURCE_NAMES = tuple(MAP_DATA_SOURCE_NAME_TO_CLASS.keys())



"""Test Manager."""
import os
import tempfile
from datetime import datetime
from pathlib import Path

import numpy as np
import pandas as pd

import nowcasting_dataset
from nowcasting_dataset.consts import SPATIAL_AND_TEMPORAL_LOCATIONS_OF_EACH_EXAMPLE_FILENAME
from nowcasting_dataset.data_sources.gsp.gsp_data_source import GSPDataSource
from nowcasting_dataset.data_sources.satellite.satellite_data_source import (
    HRVSatelliteDataSource,
    SatelliteDataSource,
)
from nowcasting_dataset.data_sources.sun.sun_data_source import SunDataSource
from nowcasting_dataset.dataset.split.split import SplitMethod
from nowcasting_dataset.manager import Manager


def test_configure_loggers():
    """Test to check loggers can be configured"""

    # set up
    manager = Manager()

    # make configuration
    local_path = Path(nowcasting_dataset.__file__).parent.parent
    filename = local_path / "tests" / "config" / "test.yaml"
    manager.load_yaml_configuration(filename=filename)

    with tempfile.TemporaryDirectory() as dst_path:

        filepath = f"{dst_path}/extra_temp_folder"
        manager.config.output_data.filepath = Path(filepath)

        manager.configure_loggers(log_level="DEBUG")


def test_sample_spatial_and_temporal_locations_for_examples():  # noqa: D103
    local_path = Path(nowcasting_dataset.__file__).parent.parent

    gsp = GSPDataSource(
        zarr_path=f"{local_path}/tests/data/gsp/test.zarr",
        start_datetime=datetime(2020, 4, 1),
        end_datetime=datetime(2020, 4, 2),
        history_minutes=30,
        forecast_minutes=60,
        image_size_pixels=64,
        meters_per_pixel=2000,
    )

    sun = SunDataSource(
        zarr_path=f"{local_path}/tests/data/sun/test.zarr",
        history_minutes=30,
        forecast_minutes=60,
    )

    manager = Manager()
    manager.data_sources = {"gsp": gsp, "sun": sun}
    manager.data_source_which_defines_geospatial_locations = gsp
    t0_datetimes = manager.get_t0_datetimes_across_all_data_sources(freq="30T")
    locations = manager.sample_spatial_and_temporal_locations_for_examples(
        t0_datetimes=t0_datetimes, n_examples=10
    )

    assert locations.columns.to_list() == ["t0_datetime_UTC", "x_center_OSGB", "y_center_OSGB"]
    assert len(locations) == 10
    assert (t0_datetimes[0] <= locations["t0_datetime_UTC"]).all()
    assert (t0_datetimes[-1] >= locations["t0_datetime_UTC"]).all()


def test_load_yaml_configuration():  # noqa: D103
    manager = Manager()
    local_path = Path(nowcasting_dataset.__file__).parent.parent
    filename = local_path / "tests" / "config" / "test.yaml"

    manager.load_yaml_configuration(filename=filename)

    manager.initialise_data_sources()
    assert len(manager.data_sources) == 8
    assert isinstance(manager.data_source_which_defines_geospatial_locations, GSPDataSource)
    assert isinstance(manager.config.process.local_temp_path, Path)


def test_get_daylight_datetime_index():
    """Check that 'manager' gets the correct t0 datetime over nighttime"""
    filename = Path(nowcasting_dataset.__file__).parent.parent / "tests" / "data" / "sat_data.zarr"

    sat = SatelliteDataSource(
        zarr_path=filename,
        history_minutes=30,
        forecast_minutes=60,
        image_size_pixels=64,
        meters_per_pixel=2000,
    )

    manager = Manager()
    manager.data_sources = {"sat": sat}
    manager.data_source_which_defines_geospatial_locations = sat
    t0_datetimes = manager.get_t0_datetimes_across_all_data_sources(freq="5T")

    # The testing sat_data.zarr has contiguous data from 12:00 to 18:00.
    # nowcasting_datamodule.history_minutes = 30
    # nowcasting_datamodule.forecast_minutes = 60
    # Daylight ends after 19:34.
    # So the expected t0_datetimes start at 12:30 (12:00 + 30 minutes)
    # and end at 17:00 (18:00 - 60 minutes)

    correct_t0_datetimes = pd.date_range("2020-04-01 12:30", "2020-04-01 17:00", freq="5 min")
    np.testing.assert_array_equal(t0_datetimes, correct_t0_datetimes)


def test_create_files_specifying_spatial_and_temporal_locations_of_each_example_if_necessary():
    """Test to initialize data sources and get batches"""

    manager = Manager()
    local_path = Path(nowcasting_dataset.__file__).parent.parent
    filename = local_path / "tests" / "config" / "test.yaml"
    manager.load_yaml_configuration(filename=filename)
    manager.config.process.n_train_batches = 10
    manager.config.process.n_validation_batches = 10
    manager.config.process.n_test_batches = 10
    manager.config.process.split_method = SplitMethod.SAME
    manager.initialise_data_sources()

    batch_size = manager.config.process.batch_size

    with tempfile.TemporaryDirectory() as local_temp_path, tempfile.TemporaryDirectory() as dst_path:  # noqa 101

        manager.config.output_data.filepath = Path(dst_path)
        manager.local_temp_path = Path(local_temp_path)

        manager.create_files_specifying_spatial_and_temporal_locations_of_each_example_if_necessary()  # noqa 101

        train_file = f"{dst_path}/train/{SPATIAL_AND_TEMPORAL_LOCATIONS_OF_EACH_EXAMPLE_FILENAME}"
        validation_file = (
            f"{dst_path}/validation/{SPATIAL_AND_TEMPORAL_LOCATIONS_OF_EACH_EXAMPLE_FILENAME}"
        )
        test_file = f"{dst_path}/test/{SPATIAL_AND_TEMPORAL_LOCATIONS_OF_EACH_EXAMPLE_FILENAME}"

        assert os.path.exists(train_file)
        assert os.path.exists(validation_file)
        assert os.path.exists(test_file)

        for file in train_file, validation_file:
            locations_df = pd.read_csv(file)
            assert len(locations_df) == 10 * batch_size

        locations_df = pd.read_csv(test_file)
        assert len(locations_df) == batch_size * manager.config.process.n_test_batches


def test_batches():
    """Test that batches can be made"""

    manager = Manager()

    # load config
    local_path = Path(nowcasting_dataset.__file__).parent.parent
    filename = local_path / "tests" / "config" / "test.yaml"
    manager.load_yaml_configuration(filename=filename)

    with tempfile.TemporaryDirectory() as local_temp_path, tempfile.TemporaryDirectory() as dst_path:  # noqa 101

        # set local temp path, and dst path
        manager.config.output_data.filepath = Path(dst_path)
        manager.local_temp_path = Path(local_temp_path)

        # set up loggers
        manager.configure_loggers(log_level="DEBUG")

        # just set satellite as data source
        filename = (
            Path(nowcasting_dataset.__file__).parent.parent / "tests" / "data" / "sat_data.zarr"
        )

        sat = SatelliteDataSource(
            zarr_path=filename,
            history_minutes=30,
            forecast_minutes=60,
            image_size_pixels=24,
            meters_per_pixel=6000,
            channels=("IR_016",),
        )

        filename = (
            Path(nowcasting_dataset.__file__).parent.parent / "tests" / "data" / "hrv_sat_data.zarr"
        )
        hrvsat = HRVSatelliteDataSource(
            zarr_path=filename,
            history_minutes=30,
            forecast_minutes=60,
            image_size_pixels=64,
            meters_per_pixel=2000,
            channels=("HRV",),
        )

        filename = (
            Path(nowcasting_dataset.__file__).parent.parent / "tests" / "data" / "gsp" / "test.zarr"
        )

        gsp = GSPDataSource(
            zarr_path=filename,
            start_datetime=datetime(2020, 4, 1),
            end_datetime=datetime(2020, 4, 2),
            history_minutes=30,
            forecast_minutes=60,
            image_size_pixels=64,
            meters_per_pixel=2000,
        )

        # Set data sources
        manager.data_sources = {"gsp": gsp, "satellite": sat, "hrvsatellite": hrvsat}
        manager.data_source_which_defines_geospatial_locations = gsp

        # make file for locations
        manager.create_files_specifying_spatial_and_temporal_locations_of_each_example_if_necessary()  # noqa 101

        # make batches
        manager.create_batches(overwrite_batches=True)

        assert os.path.exists(f"{dst_path}/train")
        assert os.path.exists(f"{dst_path}/train/gsp")
        assert os.path.exists(f"{dst_path}/train/gsp/000000.nc")
        assert os.path.exists(f"{dst_path}/train/gsp/000001.nc")
        assert os.path.exists(f"{dst_path}/train/satellite/000000.nc")
        assert os.path.exists(f"{dst_path}/train/satellite/000001.nc")
        assert os.path.exists(f"{dst_path}/train/hrvsatellite/000001.nc")
        assert os.path.exists(f"{dst_path}/train/hrvsatellite/000000.nc")

        # check logs is appended to
        for log_file in ["combined", "gsp", "satellite", "hrvsatellite"]:
            filename = f"{dst_path}/{log_file}.log"
            assert os.path.exists(filename)
            with open(filename) as f:
                num_lines = sum(1 for line in f)
                print(num_lines, log_file)
                assert num_lines > 0, f"Log {filename} is empty"


def test_save_config():
    """Test that configuration file is saved"""

    manager = Manager()

    # load config
    local_path = Path(nowcasting_dataset.__file__).parent.parent
    filename = local_path / "tests" / "config" / "test.yaml"
    manager.load_yaml_configuration(filename=filename)

    with tempfile.TemporaryDirectory() as local_temp_path, tempfile.TemporaryDirectory() as dst_path:  # noqa 101

        # set local temp path, and dst path
        manager.config.output_data.filepath = Path(dst_path)
        manager.local_temp_path = Path(local_temp_path)

        # save config
        manager.save_yaml_configuration()

        assert os.path.exists(f"{dst_path}/configuration.yaml")


def test_run():
    """Test to initialize data sources and get batches"""

    manager = Manager()
    local_path = Path(nowcasting_dataset.__file__).parent.parent
    filename = local_path / "tests" / "config" / "test.yaml"
    manager.load_yaml_configuration(filename=filename)
    manager.initialise_data_sources()

    with tempfile.TemporaryDirectory() as local_temp_path, tempfile.TemporaryDirectory() as dst_path:  # noqa 101

        manager.config.output_data.filepath = Path(dst_path)
        manager.local_temp_path = Path(local_temp_path)

        manager.create_files_specifying_spatial_and_temporal_locations_of_each_example_if_necessary()  # noqa 101
        manager.create_batches(overwrite_batches=True)



""" General Data Source output pydantic class. """
from __future__ import annotations

import logging
import os
from pathlib import Path
from typing import Tuple

import numpy as np
import xarray as xr

from nowcasting_dataset.dataset.xr_utils import PydanticXArrayDataSet
from nowcasting_dataset.filesystem.utils import makedirs
from nowcasting_dataset.utils import get_netcdf_filename

logger = logging.getLogger(__name__)


class DataSourceOutput(PydanticXArrayDataSet):
    """General Data Source output pydantic class.

    Data source output classes should inherit from this class
    """

    __slots__ = []

    def get_name(self) -> str:
        """Get the name of the class"""
        return self.__class__.__name__.lower()

    def save_netcdf(self, batch_i: int, path: Path):
        """
        Save batch to netcdf file in path/<DataSourceOutputName>/.

        Args:
            batch_i: the batch id, used to make the filename
            path: the path where it will be saved. This can be local or in the cloud.
        """
        filename = get_netcdf_filename(batch_i)

        name = self.get_name()

        # make folder
        folder = os.path.join(path, name)
        if batch_i == 0:
            # only need to make the folder once, or check that the folder is there once
            makedirs(path=folder)

        # make file
        local_filename = os.path.join(folder, filename)

        encoding = {name: {"compression": "lzf"} for name in self.data_vars}
        self.to_netcdf(local_filename, engine="h5netcdf", mode="w", encoding=encoding)

    def check_nan_and_inf(self, data: xr.Dataset, variable_name: str = None):
        """Check that all values are non NaNs and not infinite"""

        if np.isnan(data).any():
            message = f"Some {self.__class__.__name__} data values are NaNs"
            if variable_name is not None:
                message += f" ({variable_name})"
            logger.error(message)
            raise Exception(message)

        if np.isinf(data).any():
            message = f"Some {self.__class__.__name__} data values are Infinite"
            if variable_name is not None:
                message += f" ({variable_name})"
            logger.error(message)
            raise Exception(message)

    def check_nan_and_fill_warning(self, data: xr.Dataset, variable_name: str = None) -> xr.Dataset:
        """Check that all values are non NaNs and not infinite"""

        if np.isnan(data).any():
            message = f"Some {self.__class__.__name__} data values are NaNs"
            if variable_name is not None:
                message += f" ({variable_name})"
            logger.warning(message)
            data = data.fillna(0)

        return data

    def check_dataset_greater_than_or_equal_to(
        self, data: xr.Dataset, min_value: int, variable_name: str = None
    ):
        """Check data is greater than a certain value"""
        if (data < min_value).any():
            message = f"Some {self.__class__.__name__} data values are less than {min_value}"
            if variable_name is not None:
                message += f" ({variable_name})"
            logger.error(message)
            raise Exception(message)

    def check_dataset_less_than_or_equal_to(
        self, data: xr.Dataset, max_value: int, variable_name: str = None
    ):
        """Check data is less than a certain value"""
        if (data > max_value).any():
            message = f"Some {self.__class__.__name__} data values are less than {max_value}"
            if variable_name is not None:
                message += f" ({variable_name})"
            logger.error(message)
            raise Exception(message)

    def check_dataset_not_equal(
        self, data: xr.Dataset, value: int, raise_error: bool = True, variable_name: str = None
    ):
        """Check data is not equal than a certain value"""
        if np.isclose(data, value).any():
            message = f"Some {self.__class__.__name__} data values are equal to {value}"
            if variable_name is not None:
                message += f" ({variable_name})"
            if raise_error:
                logger.error(message)
                raise Exception(message)
            else:
                logger.warning(message)

    def check_data_var_dim(self, data: xr.Dataset, expected_dims: Tuple[str]):
        """Check the data var has the correct dims"""

        actual_dims = data.dims
        # check the dims are the same as expected,
        # we are using 'set' so ordering doesnt matter
        if set(actual_dims) != set(expected_dims):
            message = (
                f"Actual dims {actual_dims} does not equal {expected_dims} "
                f"for {data.name} {self.__class__.__name__}"
            )
            logger.error(message)
            raise Exception(message)


def pad_nans(array, pad_width) -> np.ndarray:
    """Pad nans with nans"""
    array = array.astype(np.float32)
    return np.pad(array, pad_width, constant_values=np.NaN)



""" Model for output of Optical Flow data """
from __future__ import annotations

import numpy as np

from nowcasting_dataset.data_sources.datasource_output import DataSourceOutput


class OpticalFlow(DataSourceOutput):
    """Class to store optical flow data as a xr.Dataset with some validation"""

    __slots__ = ()
    _expected_dimensions = ("time", "x", "y", "channels")
    _expected_data_vars = ("data",)

    @classmethod
    def model_validation(cls, v):
        """Check that all values are not NaN, Infinite, or -1."""
        assert (~np.isnan(v.data)).all(), "Some optical flow data values are NaNs"
        assert (~np.isinf(v.data)).all(), "Some optical flow data values are Infinite"
        assert (v.data != -1).all(), "Some optical flow data values are -1"
        return v



""" PV Data Source """

import datetime
import functools
import io
import logging
from dataclasses import dataclass
from numbers import Number
from pathlib import Path
from typing import List, Optional, Tuple, Union

import fsspec
import numpy as np
import pandas as pd
import xarray as xr

import nowcasting_dataset.filesystem.utils as nd_fs_utils
from nowcasting_dataset import geospatial
from nowcasting_dataset.consts import DEFAULT_N_PV_SYSTEMS_PER_EXAMPLE
from nowcasting_dataset.data_sources.data_source import ImageDataSource
from nowcasting_dataset.data_sources.pv.pv_model import PV
from nowcasting_dataset.square import get_bounding_box_mask

logger = logging.getLogger(__name__)


@dataclass
class PVDataSource(ImageDataSource):
    """PV Data Source.

    This inherits from ImageDataSource so PVDataSource can select a geospatial region of interest
    defined by image_size_pixels and meters_per_pixel.
    """

    filename: Union[str, Path]
    metadata_filename: Union[str, Path]
    # TODO: Issue #425: Use config to set start_dt and end_dt.
    start_datetime: Optional[datetime.datetime] = None
    end_datetime: Optional[datetime.datetime] = None
    random_pv_system_for_given_location: Optional[bool] = True
    #: Each example will always have this many PV systems.
    #: If less than this number exist in the data then pad with NaNs.
    n_pv_systems_per_example: int = DEFAULT_N_PV_SYSTEMS_PER_EXAMPLE
    load_azimuth_and_elevation: bool = False
    load_from_gcs: bool = True  # option to load data from gcs, or local file
    get_center: bool = True

    def __post_init__(self, image_size_pixels: int, meters_per_pixel: int):
        """Post Init"""
        super().__post_init__(image_size_pixels, meters_per_pixel)

        self.rng = np.random.default_rng()
        self.load()

    def check_input_paths_exist(self) -> None:
        """Check input paths exist.  If not, raise a FileNotFoundError."""
        for filename in [self.filename, self.metadata_filename]:
            nd_fs_utils.check_path_exists(filename)

    def load(self):
        """
        Load metadata and pv power
        """
        self._load_metadata()
        self._load_pv_power()
        self.pv_metadata, self.pv_power = align_pv_system_ids(self.pv_metadata, self.pv_power)

    @staticmethod
    def get_data_model_for_batch():
        """Get the model that is used in the batch"""
        return PV

    def _load_metadata(self):

        logger.debug(f"Loading PV metadata from {self.metadata_filename}")

        pv_metadata = pd.read_csv(self.metadata_filename, index_col="system_id")
        pv_metadata.dropna(subset=["longitude", "latitude"], how="any", inplace=True)

        pv_metadata["location_x"], pv_metadata["location_y"] = geospatial.lat_lon_to_osgb(
            pv_metadata["latitude"], pv_metadata["longitude"]
        )

        # Remove PV systems outside the geospatial boundary of the
        # satellite data:
        GEO_BOUNDARY_OSGB = {
            "WEST": -238_000,
            "EAST": 856_000,
            "NORTH": 1_222_000,
            "SOUTH": -184_000,
        }
        self.pv_metadata = pv_metadata[
            (pv_metadata.location_x >= GEO_BOUNDARY_OSGB["WEST"])
            & (pv_metadata.location_x <= GEO_BOUNDARY_OSGB["EAST"])
            & (pv_metadata.location_y <= GEO_BOUNDARY_OSGB["NORTH"])
            & (pv_metadata.location_y >= GEO_BOUNDARY_OSGB["SOUTH"])
        ]

    def _load_pv_power(self):

        logger.debug(f"Loading PV Power data from {self.filename}")

        pv_power = load_solar_pv_data(
            self.filename, start_dt=self.start_datetime, end_dt=self.end_datetime
        )

        # A bit of hand-crafted cleaning
        if 30248 in pv_power.columns:
            pv_power[30248]["2018-10-29":"2019-01-03"] = np.NaN

        # Drop columns and rows with all NaNs.
        pv_power.dropna(axis="columns", how="all", inplace=True)
        pv_power.dropna(axis="index", how="all", inplace=True)

        # drop systems with over night power
        pv_power = drop_pv_systems_which_produce_overnight(pv_power)

        # Resample to 5-minutely and interpolate up to 15 minutes ahead.
        # TODO: Issue #301: Give users the option to NOT resample (because Perceiver IO
        # doesn't need all the data to be perfectly aligned).
        pv_power = pv_power.resample("5T").interpolate(method="time", limit=3)
        pv_power.dropna(axis="index", how="all", inplace=True)
        # self.pv_power = dd.from_pandas(pv_power, npartitions=3)
        print("pv_power = {:,.1f} MB".format(pv_power.values.nbytes / 1e6))
        self.pv_power = pv_power

        # get the max generation / capacity for each system
        self.pv_capacity = pv_power.max()

    def _get_time_slice(self, t0_dt: pd.Timestamp) -> [pd.DataFrame]:
        # TODO: Cache this?
        start_dt = self._get_start_dt(t0_dt)
        end_dt = self._get_end_dt(t0_dt)
        del t0_dt  # t0 is not used in the rest of this method!
        selected_pv_power = self.pv_power.loc[start_dt:end_dt].dropna(axis="columns", how="any")
        selected_pv_capacity = self.pv_capacity[selected_pv_power.columns]

        pv_power_zero_or_above_flag = selected_pv_power.ge(0).all()

        if pv_power_zero_or_above_flag.sum() != len(selected_pv_power.columns):
            n = len(selected_pv_power.columns) - pv_power_zero_or_above_flag.sum()
            logger.debug(f"Will be removing {n} pv systems as they have negative values")

        selected_pv_power = selected_pv_power.loc[:, pv_power_zero_or_above_flag]
        selected_pv_capacity = selected_pv_capacity.loc[pv_power_zero_or_above_flag]

        return selected_pv_power, selected_pv_capacity

    def _get_central_pv_system_id(
        self,
        x_meters_center: Number,
        y_meters_center: Number,
        pv_system_ids_with_data_for_timeslice: pd.Int64Index,
    ) -> int:
        # If x_meters_center and y_meters_center have been chosen
        # by PVDataSource.pick_locations_for_batch() then we just have
        # to find the pv_system_ids at that exact location.  This is
        # super-fast (a few hundred microseconds).  We use np.isclose
        # instead of the equality operator because floats.
        pv_system_ids = self.pv_metadata.index[
            np.isclose(self.pv_metadata.location_x, x_meters_center)
            & np.isclose(self.pv_metadata.location_y, y_meters_center)
        ]
        if len(pv_system_ids) == 0:
            # TODO: Implement finding PV systems closest to x_meters_center,
            # y_meters_center.  This will probably be quite slow, so always
            # try finding an exact match first (which is super-fast).
            raise NotImplementedError(
                "Not yet implemented the ability to find PV systems *nearest*"
                " (but not at the identical location to) x_meters_center and"
                " y_meters_center."
            )

        pv_system_ids = pv_system_ids_with_data_for_timeslice.intersection(pv_system_ids)
        assert len(pv_system_ids) > 0

        # Select just one PV system (the locations in PVOutput.org are quite
        # approximate, so it's quite common to have multiple PV systems
        # at the same nominal lat lon).
        if self.random_pv_system_for_given_location:
            pv_system_id = self.rng.choice(pv_system_ids)
        else:
            pv_system_id = pv_system_ids[0]

        return pv_system_id

    def _get_all_pv_system_ids_in_roi(
        self,
        x_meters_center: Number,
        y_meters_center: Number,
        pv_system_ids_with_data_for_timeslice: pd.Int64Index,
    ) -> pd.Int64Index:
        """
        Find the PV system IDs. for all the PV systems within the geospatial

        This is for all the PV systems within the geospatial
        region of interest, defined by self.square.
        """
        logger.debug(f"Getting PV example data for {x_meters_center} and {y_meters_center}")

        bounding_box = self._square.bounding_box_centered_on(
            x_meters_center=x_meters_center, y_meters_center=y_meters_center
        )
        x = self.pv_metadata.location_x
        y = self.pv_metadata.location_y

        # make mask of pv system ids
        mask = get_bounding_box_mask(bounding_box, x, y)
        pv_system_ids = self.pv_metadata.index[mask]

        pv_system_ids = pv_system_ids_with_data_for_timeslice.intersection(pv_system_ids)

        # there may not be any pv systems in a GSP region
        # assert len(pv_system_ids) > 0

        return pv_system_ids

    def get_example(
        self, t0_dt: pd.Timestamp, x_meters_center: Number, y_meters_center: Number
    ) -> xr.Dataset:
        """
        Get Example data for PV data

        Args:
            t0_dt: list of timestamps for the datetime of the batches. The batch will also include
                data for historic and future depending on 'history_minutes' and 'future_minutes'.
            x_meters_center: x center batch locations
            y_meters_center: y center batch locations

        Returns: Example data

        """
        logger.debug("Getting PV example data")

        selected_pv_power, selected_pv_capacity = self._get_time_slice(t0_dt)
        all_pv_system_ids = self._get_all_pv_system_ids_in_roi(
            x_meters_center, y_meters_center, selected_pv_power.columns
        )
        if self.get_center:
            central_pv_system_id = self._get_central_pv_system_id(
                x_meters_center, y_meters_center, selected_pv_power.columns
            )

            # By convention, the 'target' PV system ID (the one in the center
            # of the image) must be in the first position of the returned arrays.
            all_pv_system_ids = all_pv_system_ids.drop(central_pv_system_id)
            all_pv_system_ids = all_pv_system_ids.insert(loc=0, item=central_pv_system_id)

        all_pv_system_ids = all_pv_system_ids[: self.n_pv_systems_per_example]

        selected_pv_power = selected_pv_power[all_pv_system_ids]
        selected_pv_capacity = selected_pv_capacity[all_pv_system_ids]

        pv_system_row_number = np.flatnonzero(self.pv_metadata.index.isin(all_pv_system_ids))
        pv_system_x_coords = self.pv_metadata.location_x[all_pv_system_ids]
        pv_system_y_coords = self.pv_metadata.location_y[all_pv_system_ids]

        # Save data into the PV object...

        # convert to data array
        da = xr.DataArray(
            data=selected_pv_power.values,
            dims=["time", "id"],
            coords=dict(
                id=all_pv_system_ids.values.astype(int),
                time=selected_pv_power.index.values,
            ),
        )

        capacity = xr.DataArray(
            data=selected_pv_capacity.values,
            dims=["id"],
            coords=dict(
                id=all_pv_system_ids.values.astype(int),
            ),
        )

        # convert to dataset
        pv = da.to_dataset(name="power_mw")
        pv["capacity_mwp"] = capacity

        # add pv x coords
        x_coords = xr.DataArray(
            data=pv_system_x_coords.values,
            dims=["id"],
        )

        y_coords = xr.DataArray(
            data=pv_system_y_coords.values,
            dims=["id"],
        )
        pv_system_row_number = xr.DataArray(
            data=pv_system_row_number,
            dims=["id"],
        )
        pv["x_coords"] = x_coords
        pv["y_coords"] = y_coords
        pv["pv_system_row_number"] = pv_system_row_number

        # pad out so that there are always n_pv_systems_per_example, pad with zeros
        pad_n = self.n_pv_systems_per_example - len(pv.id)
        pv = pv.pad(id=(0, pad_n), power_mw=((0, 0), (0, pad_n)), constant_values=0)

        return pv

    def get_locations(self, t0_datetimes: pd.DatetimeIndex) -> Tuple[List[Number], List[Number]]:
        """Find a valid geographical location for each t0_datetime.

        Returns:  x_locations, y_locations. Each has one entry per t0_datetime.
            Locations are in OSGB coordinates.
        """
        # Set this up as a separate function, so we can cache the result!
        @functools.cache  # functools.cache requires Python >= 3.9
        def _get_pv_system_ids(t0_datetime: pd.Timestamp) -> pd.Int64Index:
            start_dt = self._get_start_dt(t0_datetime)
            end_dt = self._get_end_dt(t0_datetime)
            available_pv_data = self.pv_power.loc[start_dt:end_dt]
            columns_mask = available_pv_data.notna().all().values
            pv_system_ids = available_pv_data.columns[columns_mask]
            assert len(pv_system_ids) > 0
            return pv_system_ids

        # Pick a random PV system for each t0_datetime, and then grab
        # their geographical location.
        x_locations = []
        y_locations = []
        for t0_datetime in t0_datetimes:
            pv_system_ids = _get_pv_system_ids(t0_datetime)
            pv_system_id = self.rng.choice(pv_system_ids)

            # Get metadata for PV system
            metadata_for_pv_system = self.pv_metadata.loc[pv_system_id]
            x_locations.append(metadata_for_pv_system.location_x)
            y_locations.append(metadata_for_pv_system.location_y)

        return x_locations, y_locations

    def datetime_index(self) -> pd.DatetimeIndex:
        """Returns a complete list of all available datetimes."""
        return self.pv_power.index


def load_solar_pv_data(
    filename: Union[str, Path],
    start_dt: Optional[datetime.datetime] = None,
    end_dt: Optional[datetime.datetime] = None,
) -> pd.DataFrame:
    """
    Load solar pv data from any compute environment.

    Args:
        filename: filename of file to be loaded
        start_dt: the start datetime, which to trim the data to
        end_dt: the end datetime, which to trim the data to

    Returns: Solar PV data
    """
    logger.debug(f"Loading Solar PV Data from {filename} from {start_dt} to {end_dt}.")

    # It is possible to simplify the code below and do
    # xr.open_dataset(file, engine='h5netcdf')
    # in the first 'with' block, and delete the second 'with' block.
    # But that takes 1 minute to load the data, where as loading into memory
    # first and then loading from memory takes 23 seconds!
    with fsspec.open(filename, mode="rb") as file:
        file_bytes = file.read()

    with io.BytesIO(file_bytes) as file:
        pv_power = xr.open_dataset(file, engine="h5netcdf")
        pv_power = pv_power.sel(datetime=slice(start_dt, end_dt))
        pv_power_df = pv_power.to_dataframe()

    # Save memory
    del file_bytes
    del pv_power

    # Process the data a little
    pv_power_df = pv_power_df.dropna(axis="columns", how="all")
    pv_power_df = pv_power_df.clip(lower=0, upper=5e7)
    # Convert the pv_system_id column names from strings to ints:
    pv_power_df.columns = [np.int32(col) for col in pv_power_df.columns]

    if "passiv" not in filename:
        pv_power_df = pv_power_df.tz_localize("Europe/London").tz_convert("UTC").tz_convert(None)

    logger.debug("Loading Solar PV Data: done")

    return pv_power_df


def align_pv_system_ids(
    pv_metadata: pd.DataFrame, pv_power: pd.DataFrame
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """Only pick PV systems for which we have metadata."""
    pv_system_ids = pv_metadata.index.intersection(pv_power.columns)
    pv_system_ids = np.sort(pv_system_ids)
    pv_power = pv_power[pv_system_ids]
    pv_metadata = pv_metadata.loc[pv_system_ids]
    return pv_metadata, pv_power


def drop_pv_systems_which_produce_overnight(pv_power: pd.DataFrame) -> pd.DataFrame:
    """Drop systems which produce power over night."""
    # TODO: Of these bad systems, 24647, 42656, 42807, 43081, 51247, 59919
    # might have some salvagable data?
    NIGHT_YIELD_THRESHOLD = 0.4
    night_hours = [22, 23, 0, 1, 2]
    pv_data_at_night = pv_power.loc[pv_power.index.hour.isin(night_hours)]
    pv_above_threshold_at_night = (pv_data_at_night > NIGHT_YIELD_THRESHOLD).any()
    bad_systems = pv_power.columns[pv_above_threshold_at_night]
    print(len(bad_systems), "bad PV systems found and removed!")
    return pv_power.drop(columns=bad_systems)



""" Square objects """
from numbers import Number
from typing import NamedTuple, Union

from nowcasting_dataset.consts import Array


class BoundingBox(NamedTuple):
    """Bounding box tuple"""

    top: Union[Number, float]
    bottom: Union[Number, float]
    left: Union[Number, float]
    right: Union[Number, float]


class Square:
    """Class for computing bounding box for satellite imagery."""

    def __init__(self, size_pixels: int, meters_per_pixel: Number):
        """
        Init

        Args:
            size_pixels: number of pixels
            meters_per_pixel: how many meters for each pixel
        """
        self.size_pixels = size_pixels
        size_meters = size_pixels * meters_per_pixel
        self._half_size_meters = size_meters / 2

    def bounding_box_centered_on(
        self, x_meters_center: Number, y_meters_center: Number
    ) -> BoundingBox:
        """
        Get bounding box from a centre

        Args:
            x_meters_center: x center of the bounding box
            y_meters_center: y center of the bounding box

        Returns: Bounding box

        """
        return BoundingBox(
            top=y_meters_center + self._half_size_meters,
            bottom=y_meters_center - self._half_size_meters,
            left=x_meters_center - self._half_size_meters,
            right=x_meters_center + self._half_size_meters,
        )


def get_bounding_box_mask(bounding_box: BoundingBox, x: Array, y: Array) -> Array:
    """
    Get boundary box mask from x and y locations. I.e are the x,y coords in the boundaring box

    Args:
        bounding_box: Bounding box
        x: x coordinates
        y: y coordinates

    Returns: list of booleans if the x and y coordinates are in the bounding box

    """
    mask = (
        (x >= bounding_box.left)
        & (x <= bounding_box.right)
        & (y >= bounding_box.bottom)
        & (y <= bounding_box.top)
    )
    return mask



""" To make fake Datasets

Wanted to keep this out of the testing frame works, as other repos, might want to use this
"""

from typing import Optional

import numpy as np
import pandas as pd
import xarray as xr

from nowcasting_dataset.config.model import Configuration
from nowcasting_dataset.consts import NWP_VARIABLE_NAMES, SAT_VARIABLE_NAMES
from nowcasting_dataset.data_sources.fake.coordinates import (
    create_random_point_coordinates_osgb,
    make_random_image_coords_osgb,
    make_random_x_and_y_osgb_centers,
)
from nowcasting_dataset.data_sources.fake.utils import (
    join_list_data_array_to_batch_dataset,
    make_t0_datetimes_utc,
)
from nowcasting_dataset.data_sources.gsp.gsp_model import GSP
from nowcasting_dataset.data_sources.metadata.metadata_model import Metadata
from nowcasting_dataset.data_sources.nwp.nwp_model import NWP
from nowcasting_dataset.data_sources.optical_flow.optical_flow_model import OpticalFlow
from nowcasting_dataset.data_sources.pv.pv_model import PV
from nowcasting_dataset.data_sources.satellite.satellite_model import HRVSatellite, Satellite
from nowcasting_dataset.data_sources.sun.sun_model import Sun
from nowcasting_dataset.data_sources.topographic.topographic_model import Topographic
from nowcasting_dataset.dataset.xr_utils import (
    convert_coordinates_to_indexes,
    convert_coordinates_to_indexes_for_list_datasets,
    join_list_dataset_to_batch_dataset,
)
from nowcasting_dataset.geospatial import lat_lon_to_osgb


def make_fake_batch(configuration: Configuration) -> dict:
    """Make fake batch"""
    batch_size = configuration.process.batch_size

    metadata = metadata_fake(batch_size=batch_size)

    return dict(
        metadata=metadata,
        satellite=satellite_fake(configuration=configuration, metadata=metadata),
        hrvsatellite=hrv_satellite_fake(
            configuration=configuration,
            metadata=metadata,
        ),
        opticalflow=optical_flow_fake(
            configuration=configuration,
            metadata=metadata,
        ),
        nwp=nwp_fake(
            configuration=configuration,
            metadata=metadata,
        ),
        pv=pv_fake(
            configuration=configuration,
            metadata=metadata,
        ),
        gsp=gsp_fake(
            configuration=configuration,
            metadata=metadata,
        ),
        sun=sun_fake(
            batch_size=batch_size,
            seq_length_5=configuration.input_data.sun.seq_length_5_minutes,
            metadata=metadata,
        ),
        topographic=topographic_fake(
            batch_size=batch_size,
            metadata=metadata,
            image_size_pixels=configuration.input_data.topographic.topographic_image_size_pixels,
        ),
    )


def gsp_fake(
    configuration: Configuration = None,
    metadata: Optional[Metadata] = None,
):
    """Create fake data"""

    if configuration is None:
        configuration = Configuration()
        configuration.input_data = Configuration().input_data.set_all_to_defaults()

    batch_size = configuration.process.batch_size
    seq_length_30 = configuration.input_data.gsp.seq_length_30_minutes
    history_seq_length = configuration.input_data.gsp.history_seq_length_30_minutes
    n_gsp_per_batch = configuration.input_data.gsp.n_gsp_per_example

    if metadata is None:
        t0_datetimes_utc = make_t0_datetimes_utc(batch_size)
        x_centers_osgb, y_centers_osgb = make_random_x_and_y_osgb_centers(batch_size)
    else:
        t0_datetimes_utc = metadata.t0_datetime_utc
        x_centers_osgb = metadata.x_center_osgb
        y_centers_osgb = metadata.y_center_osgb

    # make batch of arrays
    xr_datasets = [
        create_gsp_pv_dataset(
            seq_length=seq_length_30,
            history_seq_length=history_seq_length,
            freq="30T",
            number_of_systems=n_gsp_per_batch,
            t0_datetime_utc=t0_datetimes_utc[i],
            x_center_osgb=x_centers_osgb[i],
            y_center_osgb=y_centers_osgb[i],
            id_limit=338,
        )
        for i in range(batch_size)
    ]

    # change dimensions to dimension indexes
    xr_datasets = convert_coordinates_to_indexes_for_list_datasets(xr_datasets)

    # make dataset
    xr_dataset = join_list_dataset_to_batch_dataset(xr_datasets)

    return GSP(xr_dataset)


def metadata_fake(batch_size):
    """Make a xr dataset"""

    # get random OSGB center in the UK
    lat = np.random.uniform(51, 55, batch_size)
    lon = np.random.uniform(-2.5, 1, batch_size)
    x_centers_osgb, y_centers_osgb = lat_lon_to_osgb(lat=lat, lon=lon)

    # get random times
    t0_datetimes_utc = make_t0_datetimes_utc(batch_size)

    metadata_dict = {}
    metadata_dict["batch_size"] = batch_size
    metadata_dict["x_center_osgb"] = list(x_centers_osgb)
    metadata_dict["y_center_osgb"] = list(y_centers_osgb)
    metadata_dict["t0_datetime_utc"] = list(t0_datetimes_utc)

    return Metadata(**metadata_dict)


def nwp_fake(
    configuration: Configuration = None,
    metadata: Optional[Metadata] = None,
) -> NWP:
    """Create fake data"""

    if configuration is None:
        configuration = Configuration()
        configuration.input_data = Configuration().input_data.set_all_to_defaults()

    batch_size = configuration.process.batch_size
    image_size_pixels = configuration.input_data.nwp.nwp_image_size_pixels
    history_seq_length = configuration.input_data.nwp.history_seq_length_60_minutes
    seq_length_60 = configuration.input_data.nwp.seq_length_60_minutes
    number_nwp_channels = len(configuration.input_data.nwp.nwp_channels)

    if metadata is None:
        t0_datetimes_utc = make_t0_datetimes_utc(batch_size)
        x_centers_osgb, y_centers_osgb = make_random_x_and_y_osgb_centers(batch_size)
    else:
        t0_datetimes_utc = metadata.t0_datetime_utc
        x_centers_osgb = metadata.x_center_osgb
        y_centers_osgb = metadata.y_center_osgb

    # make batch of arrays
    xr_arrays = [
        create_image_array(
            seq_length=seq_length_60,
            history_seq_length=history_seq_length,
            image_size_pixels=image_size_pixels,
            channels=NWP_VARIABLE_NAMES[0:number_nwp_channels],
            freq="60T",
            t0_datetime_utc=t0_datetimes_utc[i],
            x_center_osgb=x_centers_osgb[i],
            y_center_osgb=y_centers_osgb[i],
        )
        for i in range(batch_size)
    ]

    # make dataset
    xr_dataset = join_list_data_array_to_batch_dataset(xr_arrays)

    xr_dataset["init_time"] = xr_dataset.time[:, 0]

    return NWP(xr_dataset)


def pv_fake(
    configuration: Configuration = None,
    metadata: Optional[Metadata] = None,
):
    """Create fake data"""

    if configuration is None:
        configuration = Configuration()
        configuration.input_data = Configuration().input_data.set_all_to_defaults()

    batch_size = configuration.process.batch_size
    seq_length_5 = configuration.input_data.pv.seq_length_5_minutes
    history_seq_length = configuration.input_data.pv.history_seq_length_5_minutes
    n_pv_systems_per_batch = configuration.input_data.pv.n_pv_systems_per_example

    if metadata is None:
        t0_datetimes_utc = make_t0_datetimes_utc(batch_size)
        x_centers_osgb, y_centers_osgb = make_random_x_and_y_osgb_centers(batch_size)
    else:
        t0_datetimes_utc = metadata.t0_datetime_utc
        x_centers_osgb = metadata.x_center_osgb
        y_centers_osgb = metadata.y_center_osgb

    # make batch of arrays
    xr_datasets = [
        create_gsp_pv_dataset(
            seq_length=seq_length_5,
            history_seq_length=history_seq_length,
            freq="5T",
            number_of_systems=n_pv_systems_per_batch,
            time_dependent_capacity=False,
            t0_datetime_utc=t0_datetimes_utc[i],
            x_center_osgb=x_centers_osgb[i],
            y_center_osgb=y_centers_osgb[i],
        )
        for i in range(batch_size)
    ]

    # change dimensions to dimension indexes
    xr_datasets = convert_coordinates_to_indexes_for_list_datasets(xr_datasets)

    # make dataset
    xr_dataset = join_list_dataset_to_batch_dataset(xr_datasets)

    return PV(xr_dataset)


def satellite_fake(
    configuration: Configuration = None,
    metadata: Optional[Metadata] = None,
) -> Satellite:
    """Create fake data"""

    if configuration is None:
        configuration = Configuration()
        configuration.input_data = Configuration().input_data.set_all_to_defaults()

    batch_size = configuration.process.batch_size
    image_size_pixels = configuration.input_data.satellite.satellite_image_size_pixels
    history_seq_length = configuration.input_data.satellite.history_seq_length_5_minutes
    seq_length_5 = configuration.input_data.satellite.seq_length_5_minutes
    number_satellite_channels = len(configuration.input_data.satellite.satellite_channels)

    if metadata is None:
        t0_datetimes_utc = make_t0_datetimes_utc(batch_size)
        x_centers_osgb, y_centers_osgb = make_random_x_and_y_osgb_centers(batch_size)
    else:
        t0_datetimes_utc = metadata.t0_datetime_utc
        x_centers_osgb = metadata.x_center_osgb
        y_centers_osgb = metadata.y_center_osgb

    # make batch of arrays
    xr_arrays = [
        create_image_array(
            seq_length=seq_length_5,
            history_seq_length=history_seq_length,
            image_size_pixels=image_size_pixels,
            channels=SAT_VARIABLE_NAMES[1:number_satellite_channels],
            t0_datetime_utc=t0_datetimes_utc[i],
            x_center_osgb=x_centers_osgb[i],
            y_center_osgb=y_centers_osgb[i],
        )
        for i in range(batch_size)
    ]

    # make dataset
    xr_dataset = join_list_data_array_to_batch_dataset(xr_arrays)

    return Satellite(xr_dataset)


def hrv_satellite_fake(
    configuration: Configuration = None,
    metadata: Optional[Metadata] = None,
) -> Satellite:
    """Create fake data"""

    if configuration is None:
        configuration = Configuration()
        configuration.input_data = Configuration().input_data.set_all_to_defaults()

    batch_size = configuration.process.batch_size
    image_size_pixels = configuration.input_data.hrvsatellite.hrvsatellite_image_size_pixels
    history_seq_length = configuration.input_data.hrvsatellite.history_seq_length_5_minutes
    seq_length_5 = configuration.input_data.hrvsatellite.seq_length_5_minutes

    if metadata is None:
        t0_datetimes_utc = make_t0_datetimes_utc(batch_size)
        x_centers_osgb, y_centers_osgb = make_random_x_and_y_osgb_centers(batch_size)
    else:
        t0_datetimes_utc = metadata.t0_datetime_utc
        x_centers_osgb = metadata.x_center_osgb
        y_centers_osgb = metadata.y_center_osgb

    # make batch of arrays
    xr_arrays = [
        create_image_array(
            seq_length=seq_length_5,
            history_seq_length=history_seq_length,
            image_size_pixels=image_size_pixels * 3,  # HRV images are 3x other images
            channels=SAT_VARIABLE_NAMES[0:1],
            t0_datetime_utc=t0_datetimes_utc[i],
            x_center_osgb=x_centers_osgb[i],
            y_center_osgb=y_centers_osgb[i],
        )
        for i in range(batch_size)
    ]

    # make dataset
    xr_dataset = join_list_data_array_to_batch_dataset(xr_arrays)

    return HRVSatellite(xr_dataset)


def optical_flow_fake(
    configuration: Configuration = None,
    metadata: Optional[Metadata] = None,
) -> OpticalFlow:
    """Create fake data"""

    if configuration is None:
        configuration = Configuration()
        configuration.input_data = Configuration().input_data.set_all_to_defaults()

    batch_size = configuration.process.batch_size
    image_size_pixels = configuration.input_data.opticalflow.opticalflow_input_image_size_pixels
    history_seq_length = configuration.input_data.opticalflow.history_seq_length_5_minutes
    seq_length_5 = configuration.input_data.opticalflow.seq_length_5_minutes
    number_satellite_channels = len(configuration.input_data.opticalflow.opticalflow_channels)

    if metadata is None:
        t0_datetimes_utc = make_t0_datetimes_utc(batch_size)
        x_centers_osgb, y_centers_osgb = make_random_x_and_y_osgb_centers(batch_size)
    else:
        t0_datetimes_utc = metadata.t0_datetime_utc
        x_centers_osgb = metadata.x_center_osgb
        y_centers_osgb = metadata.y_center_osgb

    # make batch of arrays
    xr_arrays = [
        create_image_array(
            seq_length=seq_length_5,
            history_seq_length=history_seq_length,
            freq="5T",
            image_size_pixels=image_size_pixels,
            channels=SAT_VARIABLE_NAMES[0:number_satellite_channels],
            t0_datetime_utc=t0_datetimes_utc[i],
            x_center_osgb=x_centers_osgb[i],
            y_center_osgb=y_centers_osgb[i],
        )
        for i in range(batch_size)
    ]

    # make dataset
    xr_dataset = join_list_data_array_to_batch_dataset(xr_arrays)

    return OpticalFlow(xr_dataset)


def sun_fake(
    batch_size,
    seq_length_5,
    metadata: Optional[Metadata] = None,
):
    """Create fake data"""

    if metadata is None:
        t0_datetimes_utc = make_t0_datetimes_utc(batch_size)
    else:
        t0_datetimes_utc = metadata.t0_datetime_utc

    # create dataset with both azimuth and elevation, index with time
    # make batch of arrays
    xr_arrays = [
        create_sun_dataset(seq_length=seq_length_5, t0_datetime_utc=t0_datetimes_utc[i])
        for i in range(batch_size)
    ]

    # make dataset
    xr_dataset = join_list_dataset_to_batch_dataset(xr_arrays)

    return Sun(xr_dataset)


def topographic_fake(batch_size, image_size_pixels, metadata: Optional[Metadata] = None):
    """Create fake data"""

    if metadata is None:
        x_centers_osgb, y_centers_osgb = make_random_x_and_y_osgb_centers(batch_size)
    else:
        x_centers_osgb = metadata.x_center_osgb
        y_centers_osgb = metadata.y_center_osgb

    # make batch of arrays
    xr_arrays = []
    for i in range(batch_size):

        x, y = make_random_image_coords_osgb(
            size=image_size_pixels,
            x_center_osgb=x_centers_osgb[i],
            y_center_osgb=y_centers_osgb[i],
        )

        xr_array = xr.DataArray(
            data=np.random.randn(
                image_size_pixels,
                image_size_pixels,
            ),
            dims=["x", "y"],
            coords=dict(
                x=x,
                y=y,
            ),
            name="data",
        )
        xr_arrays.append(xr_array)

    # make dataset
    xr_dataset = join_list_data_array_to_batch_dataset(xr_arrays)

    return Topographic(xr_dataset)


def create_image_array(
    dims=("time", "x", "y", "channels"),
    seq_length=19,
    history_seq_length=5,
    image_size_pixels=64,
    channels=SAT_VARIABLE_NAMES,
    freq="5T",
    t0_datetime_utc: Optional = None,
    x_center_osgb: Optional = None,
    y_center_osgb: Optional = None,
):
    """Create Satellite or NWP fake image data"""

    if t0_datetime_utc is None:
        t0_datetime_utc = make_t0_datetimes_utc(batch_size=1)[0]

    x, y = make_random_image_coords_osgb(
        size=image_size_pixels, x_center_osgb=x_center_osgb, y_center_osgb=y_center_osgb
    )

    time = pd.date_range(end=t0_datetime_utc, freq=freq, periods=history_seq_length + 1).union(
        pd.date_range(start=t0_datetime_utc, freq=freq, periods=seq_length - history_seq_length)
    )

    ALL_COORDS = {
        "time": time,
        "x": x,
        "y": y,
        "channels": np.array(channels),
    }
    coords = [(dim, ALL_COORDS[dim]) for dim in dims]
    image_data_array = xr.DataArray(
        abs(  # to make sure average is about 100
            np.random.uniform(
                0,
                200,
                size=(seq_length, image_size_pixels, image_size_pixels, len(channels)),
            )
        ),
        coords=coords,
        name="data",
    )  # Fake data for testing!

    return image_data_array


def create_gsp_pv_dataset(
    dims=("time", "id"),
    freq="5T",
    seq_length=19,
    history_seq_length=5,
    number_of_systems=128,
    time_dependent_capacity: bool = True,
    t0_datetime_utc: Optional = None,
    x_center_osgb: Optional = None,
    y_center_osgb: Optional = None,
    id_limit: int = 1000,
) -> xr.Dataset:
    """
    Create gsp or pv fake dataset

    Args:
        dims: the dims that are made for "power_mw"
        freq: the frequency of the time steps
        seq_length: the time sequence length
        number_of_systems: number of pv or gsp systems
        time_dependent_capacity: if the capacity is time dependent.
            GSP capacities increase over time,
            but PV systems are the same (or should be).
        history_seq_length: The historic length
        t0_datetime_utc: the time now, if this is not given, a random one will be made.
        x_center_osgb: the x center of the example. If not given, a random one will be made.
        y_center_osgb: the y center of the example. If not given, a random one will be made.
        id_limit: The maximum id number allowed. For example for GSP it should be 338

    Returns: xr.Dataset of fake data

    """

    if t0_datetime_utc is None:
        t0_datetime_utc = make_t0_datetimes_utc(batch_size=1)[0]

    time = pd.date_range(end=t0_datetime_utc, freq=freq, periods=history_seq_length + 1).union(
        pd.date_range(start=t0_datetime_utc, freq=freq, periods=seq_length - history_seq_length)
    )

    ALL_COORDS = {
        "time": time,
        "id": np.random.choice(range(id_limit), number_of_systems, replace=False),
    }
    coords = [(dim, ALL_COORDS[dim]) for dim in dims]

    # make pv yield.  randn samples from a Normal distribution (and so can go negative).
    # The values are clipped to be positive later.
    data = np.random.randn(seq_length, number_of_systems)

    # smooth the data, the convolution method smooths that data across systems first,
    # and then a bit across time (depending what you set N)
    N = int(seq_length / 2)
    data = np.convolve(data.ravel(), np.ones(N) / N, mode="same").reshape(
        (seq_length, number_of_systems)
    )
    # Need to clip  *after* smoothing, because the smoothing method might push
    # non-zero data below zero.  Clip at 0.1 instead of 0 so we don't get div-by-zero errors
    # if capacity is zero (capacity is computed as the max of the random numbers).
    data = data.clip(min=0.1)

    # make into a Data Array
    data_array = xr.DataArray(
        data,
        coords=coords,
    )  # Fake data for testing!

    capacity = data_array.max(dim="time")
    if time_dependent_capacity:
        capacity = capacity.expand_dims(time=seq_length)
        capacity.__setitem__("time", data_array.time.values)

    data = data_array.to_dataset(name="power_mw")

    # make random coords
    x, y = create_random_point_coordinates_osgb(
        size=number_of_systems, x_center_osgb=x_center_osgb, y_center_osgb=y_center_osgb
    )

    x_coords = xr.DataArray(
        data=x,
        dims=["id"],
    )

    y_coords = xr.DataArray(
        data=y,
        dims=["id"],
    )

    data["capacity_mwp"] = capacity
    data["x_coords"] = x_coords
    data["y_coords"] = y_coords

    # Add 1000 to the id numbers for the row numbers.
    # This is a quick way to make sure row number is different from id,
    data["pv_system_row_number"] = data["id"] + 1000

    data.__setitem__("power_mw", data.power_mw.clip(min=0))

    return data


def create_sun_dataset(
    dims=("time",),
    freq="5T",
    seq_length=19,
    t0_datetime_utc: Optional = None,
) -> xr.Dataset:
    """
    Create sun fake dataset

    Args:
        dims: # TODO
        freq: # TODO
        seq_length: # TODO
        t0_datetime_utc: # TODO

    Returns: # TODO

    """

    if t0_datetime_utc is None:
        t0_datetime_utc = make_t0_datetimes_utc(batch_size=1)[0]

    ALL_COORDS = {
        "time": pd.date_range(t0_datetime_utc, freq=freq, periods=seq_length),
    }
    coords = [(dim, ALL_COORDS[dim]) for dim in dims]
    data_array = xr.DataArray(
        np.random.randn(
            seq_length,
        ),
        coords=coords,
    )  # Fake data for testing!

    sun = data_array.to_dataset(name="elevation")
    sun["azimuth"] = sun.elevation

    sun.__setitem__("azimuth", sun.azimuth.clip(min=0, max=360))
    sun.__setitem__("elevation", sun.elevation.clip(min=-90, max=90))

    sun = convert_coordinates_to_indexes(sun)

    return sun


def create_metadata_dataset() -> xr.Dataset:
    """Create fake metadata dataset"""
    d = {
        "dims": ("t0_dt",),
        "data": pd.date_range("2021-01-01", freq="5T", periods=1) + pd.Timedelta("30T"),
    }

    data = (xr.DataArray.from_dict(d)).to_dataset(name="data")

    for v in ["x_meters_center", "y_meters_center", "object_at_center_label"]:
        d: dict = {"dims": ("t0_dt",), "data": [np.random.randint(0, 1000)]}
        d: xr.Dataset = (xr.DataArray.from_dict(d)).to_dataset(name=v)
        data[v] = getattr(d, v)

    return data


def create_datetime_dataset(
    seq_length=19,
) -> xr.Dataset:
    """Create fake datetime dataset"""
    ALL_COORDS = {
        "time": pd.date_range("2021-01-01", freq="5T", periods=seq_length),
    }
    coords = [("time", ALL_COORDS["time"])]
    data_array = xr.DataArray(
        np.random.randn(
            seq_length,
        ),
        coords=coords,
    )  # Fake data

    data = data_array.to_dataset()

    ds = data.rename({"data": "day_of_year_cos"})
    ds["day_of_year_sin"] = data.rename({"data": "day_of_year_sin"}).day_of_year_sin
    ds["hour_of_day_cos"] = data.rename({"data": "hour_of_day_cos"}).hour_of_day_cos
    ds["hour_of_day_sin"] = data.rename({"data": "hour_of_day_sin"}).hour_of_day_sin

    return data



""" utils functions """
import logging
import re
import tempfile
import threading
from concurrent import futures
from functools import wraps

import fsspec.asyn
import gcsfs
import numpy as np
import pandas as pd

import nowcasting_dataset.filesystem.utils as nd_fs_utils
from nowcasting_dataset.consts import LOG_LEVELS, Array

logger = logging.getLogger(__name__)


def set_fsspec_for_multiprocess() -> None:
    """
    Clear reference to the loop and thread.

    This is necessary otherwise
    gcsfs hangs in the ML training loop.  Only required for fsspec >= 0.9.0
    See https://github.com/dask/gcsfs/issues/379#issuecomment-839929801
    TODO: Try deleting this two lines to make sure this is still relevant.
    """
    fsspec.asyn.iothread[0] = None
    fsspec.asyn.loop[0] = None


# TODO: Issue #170. Is this this function still used?
def is_monotonically_increasing(a: Array) -> bool:
    """Check the array is monotonically increasing"""
    # TODO: Can probably replace with pd.Index.is_monotonic_increasing()
    assert a is not None
    assert len(a) > 0
    if isinstance(a, pd.DatetimeIndex):
        a = a.view(int)
    a = np.asarray(a)
    return np.all(np.diff(a) > 0)


def get_netcdf_filename(batch_idx: int) -> str:
    """Generate full filename, excluding path."""
    assert 0 <= batch_idx < 1e6
    return f"{batch_idx:06d}.nc"


class OpenData:
    """Open a file, but if from GCS, the file is downloaded to a temp file first."""

    def __init__(self, file_name):
        """Check file is there, and create temporary file"""
        self.file_name = file_name

        filesystem = nd_fs_utils.get_filesystem(file_name)
        if not filesystem.exists(file_name):
            raise RuntimeError(f"{file_name} does not exist!")

        self.temp_file = tempfile.NamedTemporaryFile()

    def __enter__(self):
        """Return filename

        1. if from gcs, download the file to temporary file, and return the temporary file name
        2. if local, return local file name
        """
        fs = nd_fs_utils.get_filesystem(self.file_name)
        if type(fs) == gcsfs.GCSFileSystem:
            fs.get_file(self.file_name, self.temp_file.name)
            filename = self.temp_file.name
        else:
            filename = self.file_name

        return filename

    def __exit__(self, type, value, traceback):
        """Close temporary file"""
        self.temp_file.close()


def remove_regex_pattern_from_keys(d: dict, pattern_to_remove: str, **regex_compile_kwargs) -> dict:
    """Remove `pattern_to_remove` from all keys in `d`.

    Return a new dict with the same values as `d`, but where the key names
    have had `pattern_to_remove` removed.
    """
    new_dict = {}
    regex = re.compile(pattern_to_remove, **regex_compile_kwargs)
    for old_key, value in d.items():
        new_key = regex.sub(string=old_key, repl="")
        new_dict[new_key] = value
    return new_dict


def configure_logger(log_level: str, logger_name: str, handlers=list[logging.Handler]) -> None:
    """Configure logger.

    Args:
      log_level: String representing logging level, e.g. 'DEBUG'.
      logger_name: String.
      handlers: A list of logging.Handler objects.
    """
    assert log_level in LOG_LEVELS
    log_level = getattr(logging, log_level)  # Convert string to int.

    formatter = logging.Formatter(
        "%(asctime)s:%(levelname)s:%(module)s#L%(lineno)d:PID=%(process)d:%(message)s"
    )

    local_logger = logging.getLogger(logger_name)
    local_logger.setLevel(log_level)

    for handler in handlers:
        handler.setLevel(log_level)
        handler.setFormatter(formatter)
        local_logger.addHandler(handler)


def get_start_and_end_example_index(batch_idx: int, batch_size: int) -> (int, int):
    """
    Get the start and end example index

    Args:
        batch_idx: the batch number
        batch_size: the size of the batches

    Returns: start and end example index

    """
    start_example_idx = batch_idx * batch_size
    end_example_idx = (batch_idx + 1) * batch_size

    return start_example_idx, end_example_idx


class DummyExecutor(futures.Executor):
    """Drop-in replacement for ThreadPoolExecutor or ProcessPoolExecutor

    This is currently not used in any code, but very useful when debugging.

    Adapted from https://stackoverflow.com/a/10436851/732596
    """

    def __init__(self, *args, **kwargs):
        """Initialise DummyExecutor."""
        self._shutdown = False
        self._shutdownLock = threading.Lock()

    def submit(self, fn, *args, **kwargs):
        """Submit task to DummyExecutor."""
        with self._shutdownLock:
            if self._shutdown:
                raise RuntimeError("cannot schedule new futures after shutdown")

            f = futures.Future()
            try:
                result = fn(*args, **kwargs)
            except BaseException as e:
                f.set_exception(e)
            else:
                f.set_result(result)

            return f

    def shutdown(self, wait=True):
        """Shutdown dummy executor."""
        with self._shutdownLock:
            self._shutdown = True


def arg_logger(func):
    """A function decorator to log all the args and kwargs passed into a function."""
    # Adapted from https://stackoverflow.com/a/23983263/732596
    @wraps(func)
    def inner_func(*args, **kwargs):
        logger.debug(f"Arguments passed into function `{func.__name__}`: {args=}; {kwargs=}")
        return func(*args, **kwargs)

    return inner_func



from datetime import timedelta

import numpy as np
import pandas as pd
import pytest

from nowcasting_dataset import time as nd_time
from nowcasting_dataset.time import FIVE_MINUTES, THIRTY_MINUTES


def test_select_daylight_datetimes():
    datetimes = pd.date_range("2020-01-01 00:00", "2020-01-02 00:00", freq="H")
    locations = [(0, 0), (20_000, 20_000)]
    daylight_datetimes = nd_time.select_daylight_datetimes(datetimes=datetimes, locations=locations)
    correct_daylight_datetimes = pd.date_range("2020-01-01 09:00", "2020-01-01 16:00", freq="H")
    np.testing.assert_array_equal(daylight_datetimes, correct_daylight_datetimes)


@pytest.mark.parametrize("min_seq_length", [2, 3, 12])
def test_get_contiguous_time_periods_1_with_1_chunk(min_seq_length):
    freq = pd.Timedelta(5, unit="minutes")
    dt_index = pd.date_range("2010-01-01", "2010-01-02", freq=freq)
    periods: pd.DataFrame = nd_time.get_contiguous_time_periods(
        dt_index, min_seq_length=min_seq_length, max_gap_duration=freq
    )
    correct_periods = pd.DataFrame([{"start_dt": dt_index[0], "end_dt": dt_index[-1]}])
    pd.testing.assert_frame_equal(periods, correct_periods)


@pytest.mark.parametrize("min_seq_length", [2, 3, 12])
def test_get_contiguous_time_periods_2_with_2_chunks(min_seq_length):
    freq = pd.Timedelta(5, unit="minutes")
    dt_index1 = pd.date_range("2010-01-01", "2010-01-02", freq=freq)
    dt_index2 = pd.date_range("2010-02-01", "2010-02-02", freq=freq)
    dt_index = dt_index1.union(dt_index2)
    periods: pd.DataFrame = nd_time.get_contiguous_time_periods(
        dt_index, min_seq_length=min_seq_length, max_gap_duration=freq
    )
    correct_periods = pd.DataFrame(
        [
            {"start_dt": dt_index1[0], "end_dt": dt_index1[-1]},
            {"start_dt": dt_index2[0], "end_dt": dt_index2[-1]},
        ]
    )
    pd.testing.assert_frame_equal(periods, correct_periods)


def test_intersection_of_2_dataframes_of_periods():
    dt = pd.Timestamp("2020-01-01 00:00")
    a = []
    b = []
    c = []  # The correct intersection

    # a: |----|       |--|
    # b:  |--|  and |------|
    # c:  |--|        |--|
    a.append({"start_dt": dt, "end_dt": dt.replace(hour=3)})
    b.append({"start_dt": dt.replace(hour=1), "end_dt": dt.replace(hour=2)})
    c.append({"start_dt": dt.replace(hour=1), "end_dt": dt.replace(hour=2)})

    # a: |---|         |---|
    # b:   |---| and |---|
    # c:   |-|         |-|
    a.append({"start_dt": dt.replace(hour=4), "end_dt": dt.replace(hour=6)})
    b.append({"start_dt": dt.replace(hour=5), "end_dt": dt.replace(hour=7)})
    c.append({"start_dt": dt.replace(hour=5), "end_dt": dt.replace(hour=6)})

    # Test identical periods:
    # a: |---|
    # b: |---|
    # c: |---|
    a.append({"start_dt": dt.replace(hour=8), "end_dt": dt.replace(hour=10)})
    b.append({"start_dt": dt.replace(hour=8), "end_dt": dt.replace(hour=10)})
    c.append({"start_dt": dt.replace(hour=8), "end_dt": dt.replace(hour=10)})

    # Test non-overlapping but adjacent periods:
    # a: |---|
    # b:     |---|
    # c:
    a.append({"start_dt": dt.replace(hour=12), "end_dt": dt.replace(hour=13)})
    b.append({"start_dt": dt.replace(hour=13), "end_dt": dt.replace(hour=14)})

    # Test non-overlapping periods:
    # a: |---|
    # b:       |---|
    # c:
    a.append({"start_dt": dt.replace(hour=15), "end_dt": dt.replace(hour=16)})
    b.append({"start_dt": dt.replace(hour=17), "end_dt": dt.replace(hour=18)})

    # Convert these lists to DataFrames:
    a = pd.DataFrame(a)
    b = pd.DataFrame(b)
    c = pd.DataFrame(c)

    # Test intersection(a, b)
    intersection = nd_time.intersection_of_2_dataframes_of_periods(a, b)
    pd.testing.assert_frame_equal(intersection, c)

    # Test intersection(b, a)
    intersection = nd_time.intersection_of_2_dataframes_of_periods(b, a)
    pd.testing.assert_frame_equal(intersection, c)

    # Test with empty DataFrames
    empty_df = pd.DataFrame(columns=["start_dt", "end_dt"])
    test_cases = [(a, empty_df), (empty_df, b), (empty_df, empty_df)]
    for test_case in test_cases:
        pd.testing.assert_frame_equal(
            nd_time.intersection_of_2_dataframes_of_periods(*test_case), empty_df
        )


def test_intersection_of_multiple_dataframes_of_periods():
    dt = pd.Timestamp("2020-01-01 00:00")

    # a: |-----|
    # b:  |---|
    # c:   |-----|
    # i:   |-|  # The correct intersection of a, b, c.
    a = pd.DataFrame([{"start_dt": dt, "end_dt": dt.replace(hour=4)}])
    b = pd.DataFrame([{"start_dt": dt.replace(hour=1), "end_dt": dt.replace(hour=3)}])
    c = pd.DataFrame([{"start_dt": dt.replace(hour=2), "end_dt": dt.replace(hour=5)}])

    i = nd_time.intersection_of_multiple_dataframes_of_periods([a, b, c])
    correct = pd.DataFrame([{"start_dt": dt.replace(hour=2), "end_dt": dt.replace(hour=3)}])
    pd.testing.assert_frame_equal(i, correct)


def test_time_periods_to_datetime_index():
    dt = pd.Timestamp("2020-01-01 00:00")
    time_periods = pd.DataFrame(
        [
            {"start_dt": dt.replace(hour=1), "end_dt": dt.replace(hour=3)},
            {"start_dt": dt.replace(hour=5), "end_dt": dt.replace(hour=10)},
        ]
    )

    FREQ = "5T"
    dt_index = nd_time.time_periods_to_datetime_index(time_periods, freq=FREQ)

    correct_dt_index = pd.date_range("2020-01-01 01:00", "2020-01-01 03:00", freq=FREQ).union(
        pd.date_range("2020-01-01 05:00", "2020-01-01 10:00", freq=FREQ)
    )
    pd.testing.assert_index_equal(dt_index, correct_dt_index)



""" Model for Topogrpahic features """

from nowcasting_dataset.data_sources.datasource_output import DataSourceOutput


class Topographic(DataSourceOutput):
    """Class to store topographic data as a xr.Dataset with some validation"""

    __slots__ = ()
    _expected_dimensions = ("x", "y")
    _expected_data_vars = ("data",)

    @classmethod
    def model_validation(cls, v):
        """Check that all values are non NaNs"""

        v.check_nan_and_inf(data=v.data)

        v.check_data_var_dim(v.data, ("example", "x_index", "y_index"))

        return v
